import xml.etree.ElementTree as ET


def _strip_xml(node):
    """Strip XML of excess whitespace.

    Source: https://stackoverflow.com/a/16919069
    """
    from xml.dom.minidom import Node

    for x in node.childNodes:
        if x.nodeType == Node.TEXT_NODE:
            if x.nodeValue:
                x.nodeValue = x.nodeValue.strip()
        elif x.nodeType == Node.ELEMENT_NODE:
            _strip_xml(x)


def make_pretty_xml(root: ET.Element) -> bytes:
    import xml.dom.minidom

    xml_doc = xml.dom.minidom.parseString(ET.tostring(root))

    _strip_xml(xml_doc)
    xml_doc.normalize()

    return xml_doc.toprettyxml(indent=" ", newl="\n", encoding="utf-8")



from __future__ import annotations

import devsim as tcad

from gdsfactory.config import logger
from gdsfactory.simulation.devsim.get_simulation_xsection import (
    PINWaveguide,
    alpha_to_k,
    dalpha_carriers,
    dn_carriers,
    k_to_alpha,
)
from gdsfactory.simulation.devsim.get_solver import DDComponent

logger.info(f"DEVSIM {tcad.__version__!r} installed at {tcad.__path__!r}")

__all__ = [
    "dn_carriers",
    "dalpha_carriers",
    "alpha_to_k",
    "k_to_alpha",
    "PINWaveguide",
    "DDComponent",
]
__version__ = "0.0.1"



from __future__ import annotations

from typing import Optional, Tuple

from numpy import floor

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.via import viac
from gdsfactory.typings import ComponentSpec, LayerSpecs


@gf.cell
def via_stack_with_offset(
    layers: LayerSpecs = ("PPP", "M1"),
    sizes: Tuple[Tuple[float, float], ...] = ((10, 10), (10, 10)),
    vias: Tuple[Optional[ComponentSpec], ...] = (None, viac),
    offsets: Optional[Tuple[float, ...]] = None,
    port_orientation: float = 180,
) -> Component:
    """Rectangular layer transition with offset between layers.

    Args:
        layers: for each via.
        vias: factory for via or None for no via.
        sizes: for each via.
        offsets: for next layer.
        port_orientation: 180: W0, 0: E0, 90: N0, 270: S0.
    """
    c = Component()
    y0 = y1 = 0

    offsets = offsets or [0] * len(layers)

    for layer, via, size, offset in zip(layers, vias, sizes, offsets):
        width, height = size
        x0 = -width / 2
        x1 = +width / 2
        y1 = y0 + height
        rect_pts = [(x0, y0), (x1, y0), (x1, y1), (x0, y1)]
        c.add_polygon(rect_pts, layer=layer)

        if via:
            via = gf.get_component(via)
            w, h = via.info["size"]
            enclosure = via.info["enclosure"]
            pitch_x, pitch_y = via.info["spacing"]

            nb_vias_x = (width - w - 2 * enclosure) / pitch_x + 1
            nb_vias_y = (height - h - 2 * enclosure) / pitch_y + 1

            nb_vias_x = int(floor(nb_vias_x)) or 1
            nb_vias_y = int(floor(nb_vias_y)) or 1

            cw = (width - (nb_vias_x - 1) * pitch_x - w) / 2
            ch = (height - (nb_vias_y - 1) * pitch_y - h) / 2

            x00 = x0 + cw + w / 2
            y00 = y0 + ch + h / 2 + offset

            ref = c.add_array(
                via, columns=nb_vias_x, rows=nb_vias_y, spacing=(pitch_x, pitch_y)
            )
            ref.move((x00, y00))

        y0 += offset
        y1 = y0 + height
        rect_pts = [(x0, y0), (x1, y0), (x1, y1), (x0, y1)]
        c.add_polygon(rect_pts, layer=layer)

    port_width = height if port_orientation in {0, 180} else width

    if port_orientation not in [0, 90, 270, 180]:
        raise ValueError(
            f"Invalid port_orientation = {port_orientation} not in [0, 90, 180, 270]"
        )
    c.add_port(
        name="e1",
        width=port_width,
        orientation=port_orientation,
        center=(0, y1),
        port_type="electrical",
        layer=list(layers)[-1],
    )
    return c


via_stack_with_offset_ppp_m1 = gf.partial(
    via_stack_with_offset,
    layers=("PPP", "M1"),
    vias=(None, viac),
)


if __name__ == "__main__":
    c = via_stack_with_offset_ppp_m1(
        layers=("SLAB90", "M1"),
        sizes=((20, 10), (20, 10)),
        vias=(viac(size=(18, 2), spacing=(5, 5)), None),
    )
    c.show(show_ports=True)



from __future__ import annotations

import pytest
from _pytest.fixtures import SubRequest

import gdsfactory as gf
from gdsfactory.config import diff_path

PDK = gf.get_generic_pdk()
PDK.activate()


@pytest.fixture
def datadir(original_datadir, tmpdir):
    return original_datadir


@pytest.fixture(scope="session")
def show_diffs(request: SubRequest) -> None:
    c = gf.read.from_gdspaths(diff_path.glob("*.gds"))
    c.show(show_ports=True)


collect_ignore = ["difftest.py"]



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight


@gf.cell
def coupler_straight_asymmetric(
    length: float = 10.0,
    gap: float = 0.27,
    width_top: float = 0.5,
    width_bot: float = 1,
    **kwargs,
) -> Component:
    """Coupler_straight_asymmetric with two parallel straights of different widths.

    Args:
        length: of straight.
        gap: between straights.
        width_top: of top straight.
        width_bot: of bottom straight.
        kwargs: cross_section settings.
    """
    component = Component()

    top = component << straight(length=length, width=width_top, **kwargs)
    bot = component << straight(length=length, width=width_bot, **kwargs)

    top.movey(0.5 * abs(width_top - width_bot) + gap + width_top)

    component.add_port("o1", port=bot.ports["o1"])
    component.add_port("o2", port=top.ports["o1"])
    component.add_port("o3", port=top.ports["o2"])
    component.add_port("o4", port=bot.ports["o2"])
    return component


if __name__ == "__main__":
    c = coupler_straight_asymmetric(length=5)
    c.show(show_ports=True)



from __future__ import annotations

from typing import List, Union

import numpy as np
from numpy import ndarray

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.geometry.functions import path_length
from gdsfactory.routing.manhattan import (
    _is_horizontal,
    _is_vertical,
    remove_flat_angles,
)
from gdsfactory.typings import (
    ComponentSpec,
    CrossSectionSpec,
    MultiCrossSectionAngleSpec,
)


def path_length_matched_points(
    list_of_waypoints: List[ndarray],
    margin: float = 0.0,
    modify_segment_i: int = -2,
    extra_length: float = 0.0,
    nb_loops: int = 1,
    bend: ComponentSpec = bend_euler,
    cross_section: Union[CrossSectionSpec, MultiCrossSectionAngleSpec] = "strip",
    **kwargs,
) -> List[ndarray]:
    """Several types of paths won't match correctly. We do not try to handle all the corner cases here. You will need to modify the input list of waypoints in some cases.

    Args:
        list_of_waypoints:  [[p1, p2, p3,...], [q1, q2, q3,...], ...]
            the number of turns have to be identical
            (usually means same number of points.
            exception is if there are some flat angles)
        margin: some extra space to budget for in addition to the bend radius
            in most cases, the default is fine
        modify_segment_i: index of the segment which accommodates the new turns
            default is next to last segment (-2)
        extra_length: distance added to all path length compensation.
            Useful is we want to add space for extra taper on all branches
        nb_loops: number of extra loops added in the path
            if nb_loops==0, no extra loop is added, instead, in each route,
            the segment indexed by `modify_segment_i` is elongated to match
            the longest route in `list_of_waypoints`
        bend: bend function
        cross_section: cross_section factory
        **kwargs

    Returns: another list of waypoints where
        - the path_length of each waypoints list are identical
        - the number of turns are identical

    """
    if nb_loops >= 1:
        return path_length_matched_points_add_waypoints(
            list_of_waypoints=list_of_waypoints,
            modify_segment_i=modify_segment_i,
            bend=bend,
            margin=margin,
            extra_length=extra_length,
            nb_loops=nb_loops,
            cross_section=cross_section,
            **kwargs,
        )
    else:
        return path_length_matched_points_modify_segment(
            list_of_waypoints=list_of_waypoints,
            modify_segment_i=modify_segment_i,
            extra_length=extra_length,
        )


def path_length_matched_points_modify_segment(
    list_of_waypoints,
    modify_segment_i,
    extra_length,
):
    if not isinstance(list_of_waypoints, list):
        raise ValueError(
            f"list_of_waypoints should be a list, got {type(list_of_waypoints)}"
        )

    list_of_waypoints = [
        remove_flat_angles(waypoints) for waypoints in list_of_waypoints
    ]
    lengths = [path_length(waypoints) for waypoints in list_of_waypoints]
    L0 = max(lengths)

    N = len(list_of_waypoints[0])

    # Find how many turns there are per path
    nb_turns = [len(waypoints) - 2 for waypoints in list_of_waypoints]

    # The paths have to have the same number of turns, otherwise this algo
    # cannot path length match
    if min(nb_turns) != max(nb_turns):
        raise ValueError(
            f"Number of turns in paths have to be identical got {nb_turns}"
        )

    if modify_segment_i < 0:
        modify_segment_i = modify_segment_i + N + 1

    list_new_waypoints = []

    # For each list of waypoints, modify one segment in-place
    for i, waypoints in enumerate(list_of_waypoints):
        p_s0, p_s1, p_next = waypoints[modify_segment_i - 1 : modify_segment_i + 2]

        p_s0 = np.array(p_s0)
        p_s1 = np.array(p_s1)

        L = lengths[i]

        # Path length compensation length
        dL = (L0 - L) / 2

        # Additional fixed length
        dL = dL + extra_length

        # Modify the segment to accommodate for path length matching
        # Two cases: vertical or horizontal segment
        if _is_vertical(p_s0, p_s1):
            sx = np.sign(p_next[0] - p_s1[0])

            dx = -sx * dL
            dp = (dx, 0)
            # Sequence of displacements to apply

        elif _is_horizontal(p_s0, p_s1):
            sy = np.sign(p_next[1] - p_s1[1])

            dy = -sy * dL
            dp = (0, dy)

        waypoints[modify_segment_i - 1] = p_s0 + dp
        waypoints[modify_segment_i] = p_s1 + dp

        list_new_waypoints += [waypoints]
    return list_new_waypoints


def path_length_matched_points_add_waypoints(
    list_of_waypoints: List[ndarray],
    modify_segment_i: int = -2,
    bend: ComponentSpec = bend_euler,
    margin: float = 0.0,
    extra_length: float = 0.0,
    nb_loops: int = 1,
    cross_section: Union[CrossSectionSpec, MultiCrossSectionAngleSpec] = "strip",
    **kwargs,
) -> List[ndarray]:
    """Args are the following.

        list_of_waypoints: a list of list_of_points:
            [[p1, p2, p3,...], [q1, q2, q3,...], ...]
            - the number of turns have to be identical
                (usually means same number of points. exception is if there are
                some flat angles)
        modify_segment_i: index of the segment which accommodates the new turns
            default is next to last segment
        bend: for bends
        margin: some extra space to budget for in addition to the bend radius
            in most cases, the default is fine
        extra_length: distance added to all path length compensation.
            Useful is we want to add space for extra taper on all branches
        nb_loops: number of extra loops added in the path
        cross_section: factory
        **kwargs: cross_section settings

    returns:
        another list of waypoints where:
            - the path_length of each waypoints list are identical
            - the number of turns are identical

    Several types of paths won't match correctly. We do not try to handle
    all the corner cases here. If the paths are not well behaved,
    the input list_of_waypoints needs to be modified.

    To have flexibility in the path length, we need to add 4 bends
    One path has to be converted in this way:

    .. code::

             __
            |  |
            |  |  This length is adjusted to make all path with the same length
            |  |
         ___|  |___

    """
    if not isinstance(list_of_waypoints, list):
        raise ValueError(
            f"list_of_waypoints should be a list, got {type(list_of_waypoints)}"
        )
    list_of_waypoints = [
        remove_flat_angles(waypoints) for waypoints in list_of_waypoints
    ]
    lengths = [path_length(waypoints) for waypoints in list_of_waypoints]
    L0 = max(lengths)
    N = len(list_of_waypoints[0])

    # Find how many turns there are per path
    nb_turns = [len(waypoints) - 2 for waypoints in list_of_waypoints]

    # The paths have to have the same number of turns, otherwise cannot path-length
    # match with this algorithm
    if min(nb_turns) != max(nb_turns):
        raise ValueError(
            f"Number of turns in paths have to be identical got {nb_turns}"
        )

    # Get the points for the segment we need to modify
    bend90 = gf.get_component(bend, cross_section=cross_section, **kwargs)

    a = margin + bend90.info["dy"]
    if modify_segment_i < 0:
        modify_segment_i = modify_segment_i + N + 1
    list_new_waypoints = []

    for i, waypoints in enumerate(list_of_waypoints):
        p_s0, p_s1, p_next = waypoints[modify_segment_i - 2 : modify_segment_i + 1]
        p_s1 = np.array(p_s1)
        L = lengths[i]

        # Path length compensation length
        dL = (L0 - L) / (2 * nb_loops)

        # Additional fixed length
        dL = dL + extra_length

        # Generate a new sequence of points which will replace this segment
        # Two cases: vertical or horizontal segment
        if _is_vertical(p_s0, p_s1):
            sx = np.sign(p_next[0] - p_s1[0])
            sy = np.sign(p_s1[1] - p_s0[1])

            dx = sx * (2 * a + dL)
            dy = sy * 2 * a

            # First new point to insert
            q0 = p_s1 + (0, -2 * nb_loops * dy)

            # Sequence of displacements to apply
            seq = [(dx, 0), (0, dy), (-dx, 0), (0, dy)] * nb_loops
            seq.pop()  # Remove last point to avoid flat angle with next point

        elif _is_horizontal(p_s0, p_s1):
            sy = np.sign(p_next[1] - p_s1[1])
            sx = np.sign(p_s1[0] - p_s0[0])

            dx = sx * 2 * a
            dy = sy * (2 * a + dL)

            # First new point to insert
            q0 = p_s1 + (-2 * dx * nb_loops, 0)

            # Sequence of displacements to apply
            seq = [(0, dy), (dx, 0), (0, -dy), (dx, 0)] * nb_loops
            seq.pop()  # Remove last point to avoid flat angle with next point

        # Generate points to insert
        qs = [q0]
        for dp in seq:
            qs += [qs[-1] + dp]

        inserted_points = np.stack(qs, axis=0)
        waypoints = np.array(waypoints)

        # Insert the points
        new_points = np.vstack(
            [
                waypoints[: modify_segment_i - 1],
                inserted_points,
                waypoints[modify_segment_i - 1 :],
            ]
        )
        list_new_waypoints += [new_points]

    return list_new_waypoints


if __name__ == "__main__":
    c = gf.Component()
    c1 = c << gf.components.straight_array(n=4, spacing=50)
    c2 = c << gf.components.straight_array(n=4, spacing=20)
    c1.y = 0
    c2.y = 0
    c2.x = 300

    routes = gf.routing.get_bundle_path_length_match(
        c1.get_ports_list(prefix="E"),
        c2.get_ports_list(prefix="W"),
        radius=5,
        layer=(2, 0),
    )
    for route in routes:
        c.add(route.references)
    c.show(show_ports=True)



"""Command line interface. Type `gf` into a terminal."""

from __future__ import annotations

import pathlib
from typing import Optional

import click
from click.core import Context, Option

import gdsfactory
from gdsfactory.config import cwd, print_config
from gdsfactory.config import print_version as _print_version
from gdsfactory.generic_tech import LAYER
from gdsfactory.install import install_gdsdiff, install_klayout_package
from gdsfactory.technology import lyp_to_dataclass
from gdsfactory.typings import PathType
from gdsfactory.write_cells import write_cells as write_cells_to_separate_gds

VERSION = "6.53.0"
LAYER_LABEL = LAYER.LABEL


def print_version(ctx: Context, param: Option, value: bool) -> None:
    """Prints the version."""
    if not value or ctx.resilient_parsing:
        return
    _print_version()
    ctx.exit()


# TOOL
@click.group()
def tool() -> None:
    """Commands working with gdsfactory tool."""
    pass


@click.command(name="config")
@click.argument("key", required=False, default=None)
def config_get(key: str) -> None:
    """Shows key values from CONFIG."""
    print_config(key)


# GDS


@click.group()
def gds() -> None:
    """Commands for dealing with GDS files."""
    pass


@click.command(name="layermap_to_dataclass")
@click.argument("filepath", type=click.Path(exists=True))
@click.option("--force", "-f", default=False, help="Force deletion", is_flag=True)
def layermap_to_dataclass(filepath, force: bool) -> None:
    """Converts KLayout LYP to a dataclass."""
    filepath_lyp = pathlib.Path(filepath)
    filepath_py = filepath_lyp.with_suffix(".py")
    if not filepath_lyp.exists():
        raise FileNotFoundError(f"{filepath_lyp} not found")
    if not force and filepath_py.exists():
        raise FileExistsError(f"found {filepath_py}")
    lyp_to_dataclass(lyp_filepath=filepath_lyp)


@click.command(name="write_cells")
@click.argument("gdspath")
@click.argument("dirpath", required=False, default=None)
def write_cells(gdspath, dirpath) -> None:
    """Write each all level cells into separate GDS files."""
    write_cells_to_separate_gds(gdspath=gdspath, dirpath=dirpath)


@click.command(name="merge_gds")
@click.argument("dirpath", required=False, default=None)
@click.argument("gdspath", required=False, default=None)
def merge_gds(
    dirpath: Optional[PathType] = None, gdspath: Optional[PathType] = None
) -> None:
    """Merges GDS cells from a directory into a single GDS."""
    dirpath = dirpath or pathlib.Path.cwd()
    gdspath = gdspath or pathlib.Path.cwd() / "merged.gds"

    dirpath = pathlib.Path(dirpath)

    c = gdsfactory.read.from_gdsdir(dirpath=dirpath)
    c.write_gds(gdspath=gdspath)
    c.show(show_ports=True)


# @click.group()
# def watch() -> None:
#     """Watch YAML or python files."""
#     pass
# @click.option("--debug", "-d", default=False, help="debug", is_flag=True)
# @click.command()
# def webapp(debug: bool = False) -> None:
#     """Opens YAML based webapp."""
#     from gdsfactory.icyaml import app

#     if debug:
#         app.run_debug()

#     else:
#         app.run()


@click.argument("path", type=click.Path(exists=True), required=False, default=cwd)
@click.command()
def watch(path=cwd) -> None:
    """Filewatch a folder for changes in python or pic.yaml files."""
    from gdsfactory.watch import watch

    path = pathlib.Path(path)
    path = path.parent if path.is_dir() else path
    watch(str(path))


# EXTRA
@click.command()
@click.argument("filename")
def show(filename: str) -> None:
    """Show a GDS file using klive."""
    gdsfactory.show(filename)


@click.command()
@click.argument("gdspath1")
@click.argument("gdspath2")
@click.option("--xor", "-x", default=False, help="include xor", is_flag=True)
def diff(gdspath1: str, gdspath2: str, xor: bool = False) -> None:
    """Show boolean difference between two GDS files."""
    from gdsfactory.gdsdiff.gdsdiff import gdsdiff

    diff = gdsdiff(gdspath1, gdspath2, xor=xor)
    diff.show()


@click.command()
def install() -> None:
    """Install Klive and generic tech layermap."""
    install_klayout_package()
    install_gdsdiff()


@click.group()
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Show the version number.",
)
def cli(name="gf") -> None:
    """`gf` is the gdsfactory command line tool."""


gds.add_command(layermap_to_dataclass)
gds.add_command(write_cells)
gds.add_command(merge_gds)
gds.add_command(show)
gds.add_command(diff)

tool.add_command(config_get)
tool.add_command(install)

# yaml.add_command(webapp)
# watch.add_command(watch_yaml)

cli.add_command(gds)
cli.add_command(tool)
cli.add_command(watch)


if __name__ == "__main__":
    # cli()
    print_version()



"""Dummy fill to keep density constant.

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""
from __future__ import annotations

import itertools
from typing import List, Optional, Union

import gdstk
import numpy as np
from numpy import sqrt

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.component_layout import _parse_layer
from gdsfactory.components.rectangle import rectangle
from gdsfactory.typings import Float2, Floats, LayerSpecs


def _loop_over(var):
    """Checks if a variable is in the form of an iterable (list/tuple)\
    and if not, returns it as a list.

    Useful for allowing argument inputs to be either lists
    (e.g. [1, 3, 4]) or single-valued (e.g. 3).
    Returns list Variable converted to list if single-valued input.

    Args:
        var : int or float or list
            Variable to check for iterability.
    """
    return var if hasattr(var, "__iter__") else [var]


def _rasterize_polygons(polygons, bounds=([-100, -100], [100, 100]), dx=1, dy=1):
    """Converts polygons to a black/white (1/0) matrix."""
    try:
        from skimage import draw
    except ImportError as e:
        raise ImportError(
            "The fill function requires the module "
            '"scikit-image" to operate.  Please retry '
            "after installing scikit-image:\n\n"
            "$ pip install --upgrade scikit-image"
        ) from e

    # Prepare polygon array by shifting all points into the first quadrant and
    # separating points into x and y lists
    xpts = []
    ypts = []
    for p in polygons:
        p_array = np.asarray(p)
        x = p_array[:, 0]
        y = p_array[:, 1]
        xpts.append((x - bounds[0][0]) / dx - 0.5)
        ypts.append((y - bounds[0][1]) / dy - 0.5)

    # Initialize the raster matrix we'll be writing to
    xsize = int(np.ceil(bounds[1][0] - bounds[0][0]) / dx)
    ysize = int(np.ceil(bounds[1][1] - bounds[0][1]) / dy)
    raster = np.zeros((ysize, xsize), dtype=bool)

    # TODO: Replace polygon_perimeter with the supercover version
    for n in range(len(xpts)):
        rr, cc = draw.polygon(ypts[n], xpts[n], shape=raster.shape)
        rrp, ccp = draw.polygon_perimeter(
            ypts[n], xpts[n], shape=raster.shape, clip=False
        )
        raster[rr, cc] = 1
        raster[rrp, ccp] = 1

    return raster


def _raster_index_to_coords(i, j, bounds=([-100, -100], [100, 100]), dx=1, dy=1):
    """Converts (i,j) index of raster matrix to real coordinates."""
    x = (j + 0.5) * dx + bounds[0][0]
    y = (i + 0.5) * dy + bounds[0][1]
    return x, y


def _expand_raster(raster, distance=(4, 2)):
    """Expands all black (1) pixels in the raster."""
    try:
        from skimage import draw, morphology
    except ImportError as e:
        raise ImportError(
            "The fill function requires the module "
            '"scikit-image" to operate.  Please retry '
            "after installing scikit-image:\n\n"
            "$ pip install --upgrade scikit-image"
        ) from e
    if distance[0] <= 0.5 and distance[1] <= 0.5:
        return raster

    num_pixels = np.array(np.ceil(distance), dtype=int)
    neighborhood = np.zeros((num_pixels[1] * 2 + 1, num_pixels[0] * 2 + 1), dtype=bool)
    rr, cc = draw.ellipse(
        num_pixels[1], num_pixels[0], distance[1] + 0.5, distance[0] + 0.5
    )
    neighborhood[rr, cc] = 1

    return morphology.binary_dilation(image=raster, footprint=neighborhood)


@cell
def fill_cell_rectangle(
    size: Float2 = (20.0, 20.0),
    layers: LayerSpecs = (0, 1, 3),
    densities: Floats = (0.5, 0.25, 0.7),
    inverted=(False, False, False),
) -> Component:
    """Returns Component on multiple layers to be used as fill.

    based on phidl.geometry

    Args:
        size: x, y dimensions of the fill area for all layers.
        layers: Specific layer(s) to put fill cell rectangle geometry on.
        densities: Fill densities for each layer specified in ``layers``.
            Must be the same size as ``layers``.
        inverted: array-like or bool
            If true, inverts the fill area for corresponding layer.
            Must be the same size as ``layers``.

    """
    D = Component()
    for layer, density, inv in zip(layers, densities, inverted):
        rectangle_size = np.array(size) * sqrt(density)
        # r = D.add_ref(rectangle(size = rectangle_size, layer = layer))
        R = rectangle(size=tuple(rectangle_size), layer=layer, port_type=None)
        R.center = (0, 0)

        if inv is True:
            A = rectangle(size=size)
            A.center = (0, 0)
            A = A.get_polygons()
            B = R.get_polygons()
            p = gdstk.boolean(A, B, operation="not")
            D.add_polygon(p, layer=layer)
        else:
            D.add_ref(R)
    return D


@cell
def fill_rectangle(
    component: Component,
    fill_layers: LayerSpecs,
    fill_size: Float2 = (5.0, 5.0),
    avoid_layers: LayerSpecs = None,
    include_layers: LayerSpecs = None,
    margin: float = 5.0,
    fill_densities: Union[float, Floats] = (0.5, 0.25, 0.7),
    fill_inverted: Optional[List[float]] = None,
    bbox: Optional[Float2] = None,
) -> Component:
    """Returns rectangular fill pattern and fills all empty areas.

    In the input component and returns a component that contains just the fill
    Dummy fill keeps density constant during fabrication

    Args:
        component: Component to fill.
        fill_layers: list of layers. fill pattern layers.
        fill_size: Rectangular size of the fill element.
        avoid_layers: Layers to be avoided (not filled) in D.
        include_layers: Layers to be filled, supersedes avoid_layers.
        margin: Margin spacing around avoided areas.
        fill_densities: defines the fill pattern density (1.0 == fully filled).
        fill_inverted: inverts the fill pattern.
        bbox: x, y limit the fill pattern to the area defined by this bounding box.

    """
    D = component

    # Create the fill cell.
    # If fill_inverted is not specified, assume all False
    fill_layers = _loop_over(fill_layers)
    fill_densities = _loop_over(fill_densities)
    if fill_inverted is None:
        fill_inverted = [False] * len(fill_layers)

    fill_inverted = _loop_over(fill_inverted)
    if len(fill_layers) != len(fill_densities):
        raise ValueError(
            "fill_rectangle() `fill_layers` and `fill_densities` parameters "
            "must be lists of the same length"
        )
    if len(fill_layers) != len(fill_inverted):
        raise ValueError(
            "fill_rectangle() `fill_layers` and `fill_inverted` parameters must "
            "be lists of the same length"
        )

    fill_cell = fill_cell_rectangle(
        size=fill_size,
        layers=fill_layers,
        densities=fill_densities,
        inverted=fill_inverted,
    )
    F = Component()

    avoid_layers = [_parse_layer(layer) for layer in _loop_over(avoid_layers)]
    exclude_polys = D.get_polygons(by_spec=True, depth=None, as_array=False)

    if avoid_layers:
        exclude_polys = {
            key: exclude_polys[key] for key in exclude_polys if key in avoid_layers
        }

    exclude_polys = [
        polygon.points for polygons in exclude_polys.values() for polygon in polygons
    ]

    if include_layers is None:
        include_polys = []
    else:
        include_layers = [_parse_layer(layer) for layer in _loop_over(include_layers)]
        include_polys = D.get_polygons(by_spec=True, depth=None)
        include_polys = {
            key: include_polys[key] for key in include_polys if key in include_layers
        }
        include_polys = itertools.chain.from_iterable(include_polys.values())

    if bbox is None:
        bbox = D.bbox

    raster = _rasterize_polygons(
        polygons=exclude_polys, bounds=bbox, dx=fill_size[0], dy=fill_size[1]
    )
    raster = raster & ~_rasterize_polygons(
        polygons=include_polys, bounds=bbox, dx=fill_size[0], dy=fill_size[1]
    )
    raster = _expand_raster(raster, distance=margin / np.array(fill_size))

    for i in range(np.size(raster, 0)):
        sub_rasters = [list(g) for k, g in itertools.groupby(raster[i])]
        j = 0
        for s in sub_rasters:
            if s[0] == 0:
                x, y = _raster_index_to_coords(i, j, bbox, fill_size[0], fill_size[1])
                a = F.add_array(fill_cell, columns=len(s), rows=1, spacing=fill_size)
                a.move((x, y))
            j += len(s)
    return F


def test_fill():
    import gdsfactory as gf
    from gdsfactory.difftest import difftest

    c = gf.Component("test_fill")
    wg = c << gf.components.straight()
    c << gf.add_padding_container(wg.parent, default=15)
    fill = fill_rectangle(
        c,
        fill_layers=((2, 0),),
        fill_densities=(1.0,),
        # fill_densities=0.5,
        avoid_layers=((1, 0),),
        # bbox=(100.0, 100.0),
    )
    c << fill
    difftest(c)
    return c


if __name__ == "__main__":
    # c = test_fill()
    # c.show()

    import gdsfactory as gf

    # c = gf.components.straight()
    # c = gf.add_padding_container(c, default=15)
    # c << fill_rectangle(
    #     c,
    #     fill_layers=((2, 0),),
    #     # fill_densities=(1.0,),
    #     # fill_densities=0.5,
    #     # avoid_layers=((1, 0),),
    #     # bbox=(100.0, 100.0),
    # )
    # c.show(show_ports=True)
    # import gdsfactory as gf
    # coupler_lengths = [10, 20, 30, 40, 50, 60, 70, 80]
    # coupler_gaps = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]
    # delta_lengths = [10, 100, 200, 300, 400, 500, 500]
    # mzi = gf.components.mzi_lattice(
    #     coupler_lengths=coupler_lengths,
    #     coupler_gaps=coupler_gaps,
    #     delta_lengths=delta_lengths,
    # )
    # # Add fill
    mzi = gf.components.mzi()
    c = gf.Component("component_with_fill")
    layers = [(1, 0)]
    fill_size = [0.5, 0.5]

    c << gf.fill_rectangle(
        mzi,
        fill_size=fill_size,
        fill_layers=layers,
        margin=5,
        fill_densities=[0.8] * len(layers),
        avoid_layers=layers,
    )
    c << mzi
    c.show(show_ports=True)



"""Gdsfactory loads configuration from 3 files, high priority overwrites low.

priority:

1. A config.yml found in the current working directory (highest priority)
2. ~/.gdsfactory/config.yml specific for the machine
3. the yamlpath_default in gdsfactory.technology.yml (lowest priority)

You can access the CONF dictionary with `print_config`

PATH has all your computer specific paths that we do not care to store

"""

from __future__ import annotations

import warnings
import sys
import io
import json
import os
import pathlib
import subprocess
from pathlib import Path
from pprint import pprint
from typing import Any, Iterable, Optional, Union
import importlib

from loguru import logger
import omegaconf
from omegaconf import OmegaConf

from rich.console import Console
from rich.table import Table

__version__ = "6.53.0"
PathType = Union[str, pathlib.Path]

home = pathlib.Path.home()
cwd = pathlib.Path.cwd()
module_path = pathlib.Path(__file__).parent.absolute()
repo_path = module_path.parent
home_path = pathlib.Path.home() / ".gdsfactory"
diff_path = repo_path / "gds_diff"
logpath = home_path / "log.log"

yamlpath_cwd = cwd / "config.yml"
yamlpath_default = module_path / "config.yml"
yamlpath_home = home_path / "config.yml"

MAX_NAME_LENGTH = 32

logger.remove()
logger.add(sink=sys.stderr, level="WARNING")

showwarning_ = warnings.showwarning


def showwarning(message, *args, **kwargs):
    logger.warning(message)
    showwarning_(message, *args, **kwargs)


warnings.showwarning = showwarning

plugins = ["ray", "femwell", "devsim", "tidy3d", "meep", "meow", "lumapi", "sax"]
pdks = ["gf45", "tj", "imec", "amf", "sky130", "ubcpdk", "aim", "ct"]


def print_version():
    """Print gdsfactory plugin versions and paths."""
    table = Table(title="Modules")
    table.add_column("Package", justify="right", style="cyan", no_wrap=True)
    table.add_column("version", style="magenta")
    table.add_column("Path", justify="right", style="green")

    table.add_row("gdsfactory", __version__, str(module_path))

    for plugin in plugins:
        try:
            m = importlib.import_module(plugin)
            try:
                table.add_row(plugin, str(m.__version__), str(m.__path__))
            except AttributeError:
                table.add_row(plugin, "", "")
        except ImportError:
            table.add_row(plugin, "not installed", "")

    console = Console()
    console.print(table)


def print_version_pdks():
    """Print gdsfactory PDK versions and paths."""
    table = Table(title="PDKs")
    table.add_column("Package", justify="right", style="cyan", no_wrap=True)
    table.add_column("version", style="magenta")
    table.add_column("Path", justify="right", style="green")

    for pdk in pdks:
        try:
            m = importlib.import_module(pdk)
            try:
                table.add_row(pdk, str(m.__version__), str(m.__path__))
            except AttributeError:
                table.add_row(pdk, "", "")
        except ImportError:
            table.add_row(pdk, "not installed", "")

    console = Console()
    console.print(table)


default_config = io.StringIO(
    """
plotter: matplotlib
sparameters_path: ${oc.env:HOME}/.gdsfactory/sparameters/generic
show_ports: True
"""
)


def set_log_level(level: str, sink=sys.stderr) -> None:
    """Sets log level for gdsfactory.

    Args:
        level: ["DEBUG", "INFO", "WARNING", "ERROR"]
        sink: defaults to standard error.
    """
    log_levels = ["DEBUG", "INFO", "WARNING", "ERROR"]
    if level not in log_levels:
        raise ValueError(f"{level!r} not a valid log level {log_levels}")
    logger.remove()
    logger.add(sink=sink, level=level)


class Paths:
    module = module_path
    repo = repo_path
    results_tidy3d = home / ".tidy3d"
    generic_tech = module / "generic_tech"
    klayout = generic_tech / "klayout"
    klayout_tech = klayout / "tech"
    klayout_lyp = klayout_tech / "layers.lyp"
    klayout_yaml = generic_tech / "layer_views.yaml"
    schema_netlist = repo_path / "tests" / "schemas" / "netlist.json"
    netlists = module_path / "samples" / "netlists"
    gdsdir = repo_path / "tests" / "gds"
    gdslib = repo_path / "gdslib"
    modes = gdslib / "modes"
    gdsdiff = gdslib / "gds"
    sparameters = gdslib / "sp"
    interconnect = gdslib / "interconnect"
    optimiser = repo_path / "tune"
    notebooks = repo_path / "docs" / "notebooks"


def read_config(
    yamlpaths: Iterable[PathType] = (yamlpath_default, yamlpath_home, yamlpath_cwd),
) -> omegaconf.DictConfig:
    config = OmegaConf.load(default_config)
    for yamlpath in set(yamlpaths):
        yamlpath = pathlib.Path(yamlpath)
        if os.access(yamlpath, os.R_OK) and yamlpath.exists():
            logger.info(f"loading tech config from {yamlpath}")
            config_new = OmegaConf.load(yamlpath)
            config = OmegaConf.merge(config, config_new)
    return config


CONF = read_config()
PATH = Paths()
sparameters_path = PATH.sparameters


def rich_output() -> None:
    """Enables rich output."""
    try:
        from rich import pretty

        pretty.install()
    except ImportError:
        print("You can install `pip install gdsfactory[full]` for better visualization")


def complex_encoder(z):
    if isinstance(z, pathlib.Path):
        return str(z)
    elif callable(z):
        return str(z.__name__)
    else:
        type_name = type(z)
        raise TypeError(f"Object {z} of type {type_name} is not serializable")


def write_config(config: Any, json_out_path: Path) -> None:
    """Write config to a JSON file."""
    with open(json_out_path, "w") as f:
        json.dump(config, f, indent=2, sort_keys=True, default=complex_encoder)


def print_config(key: Optional[str] = None) -> None:
    """Prints a key for the config or all the keys."""
    if key:
        if CONF.get(key):
            print(CONF[key])
        else:
            print(f"{key!r} key not found in {CONF.keys()}")
    else:
        pprint(CONF)


def call_if_func(f: Any, **kwargs) -> Any:
    """Calls function if it's a function Useful to create objects from.

    functions if it's an object it just returns the object.
    """
    return f(**kwargs) if callable(f) else f


def get_git_hash():
    """Returns repository git hash."""
    try:
        with open(os.devnull, "w") as shutup:
            return (
                subprocess.check_output(["git", "rev-parse", "HEAD"], stderr=shutup)
                .decode("utf-8")
                .strip("\n")
            )
    except subprocess.CalledProcessError:
        return "not_a_git_repo"


def set_plot_options(
    show_ports: bool = True,
    show_subports: bool = False,
    label_aliases: bool = False,
    new_window: bool = False,
    blocking: bool = False,
    zoom_factor: float = 1.4,
) -> None:
    """Set plot options for matplotlib."""
    from gdsfactory.quickplotter import set_quickplot_options

    set_quickplot_options(
        show_ports=show_ports,
        show_subports=show_subports,
        label_aliases=label_aliases,
        new_window=new_window,
        blocking=blocking,
        zoom_factor=zoom_factor,
    )


if __name__ == "__main__":
    # print(PATH.sparameters)
    # print_config()
    print_version()
    # write_tech("tech.json")



from __future__ import annotations

import pathlib

import gdsfactory as gf
from gdsfactory.simulation.modes.find_neff_vs_width import find_neff_vs_width

PATH = pathlib.Path(__file__).parent.absolute() / "modes"

nm = 1e-3
wg_thickness_silicon = 220 * nm
wg_thickness_nitride = 400 * nm


find_neff_vs_width_strip = gf.partial(
    find_neff_vs_width,
    wg_thickness=wg_thickness_silicon,
    slab_thickness=0,
    filepath=PATH / "neff_vs_width_strip.csv",
)
find_neff_vs_width_rib90 = gf.partial(
    find_neff_vs_width,
    wg_thickness=wg_thickness_silicon,
    slab_thickness=90 * nm,
    filepath=PATH / "neff_vs_width_rib.csv",
)

find_neff_vs_width_nitride = gf.partial(
    find_neff_vs_width,
    wg_thickness=wg_thickness_nitride,
    slab_thickness=0.0,
    ncore=2.0,
    width1=200 * nm,
    width2=1500 * nm,
    filepath=PATH / "neff_vs_width_nitride.csv",
)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    import gdsfactory.simulation.modes as gm

    df = find_neff_vs_width_strip()
    df = find_neff_vs_width_rib90()
    df = find_neff_vs_width_nitride()
    gm.plot_neff_vs_width(df)
    plt.show()



import numpy as np
from gdsfactory.simulation.devsim import get_simulation_xsection
import gdsfactory as gf

PDK = gf.get_generic_pdk()
PDK.activate()


def test_pin_waveguide():
    """Test reverse bias waveguide."""
    nm = 1e-9
    c = get_simulation_xsection.PINWaveguide(
        wg_width=500 * nm,
        wg_thickness=220 * nm,
        slab_thickness=90 * nm,
    )
    c.ddsolver()
    n_dist = {}
    neffs = {}

    vmin = 0
    vmax = -5
    voltages = [vmin, vmax]
    Vstep = vmax

    for ind, voltage in enumerate(voltages):
        Vinit = 0 if ind == 0 else voltages[ind - 1]
        c.ramp_voltage(Vfinal=voltage, Vstep=Vstep, Vinit=Vinit)
        waveguide = c.make_waveguide(wavelength=1.55)
        waveguide.compute_modes(isolate=True)
        n_dist[voltage] = waveguide.nx
        neffs[voltage] = waveguide.neffs[0]

    dn = neffs[vmin] - neffs[vmax]
    assert np.isclose(dn.real, -0.00011342135795189279), dn.real


if __name__ == "__main__":
    nm = 1e-9
    c = get_simulation_xsection.PINWaveguide(
        wg_width=500 * nm,
        wg_thickness=220 * nm,
        slab_thickness=90 * nm,
    )
    c.ddsolver()
    n_dist = {}
    neffs = {}

    vmin = 0
    vmax = -5
    voltages = [vmin, vmax]
    Vstep = vmax

    for ind, voltage in enumerate(voltages):
        Vinit = 0 if ind == 0 else voltages[ind - 1]
        c.ramp_voltage(Vfinal=voltage, Vstep=Vstep, Vinit=Vinit)
        waveguide = c.make_waveguide(wavelength=1.55)
        waveguide.compute_modes(isolate=True)
        n_dist[voltage] = waveguide.nx
        neffs[voltage] = waveguide.neffs[0]

    dn = neffs[vmin] - neffs[vmax]



from __future__ import annotations

from functools import partial
from typing import Optional

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, LayerSpec

# The default values are loosely based on Taillaert et al,
#  "A Compact Two-Dimensional Grating Coupler Used
# as a Polarization Splitter", IEEE Phot. Techn. Lett. 15(9), 2003

rectangle_unit_cell = partial(
    rectangle, size=(0.3, 0.3), layer="SLAB150", centered=True, port_type=None
)


@gf.cell
def grating_coupler_dual_pol(
    unit_cell: ComponentSpec = rectangle_unit_cell,
    period_x: float = 0.58,
    period_y: float = 0.58,
    x_span: float = 11,
    y_span: float = 11,
    length_taper: float = 150.0,
    width_taper: float = 10.0,
    polarization: str = "dual",
    wavelength: float = 1.55,
    taper: ComponentSpec = taper_function,
    base_layer: Optional[LayerSpec] = "WG",
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    r"""2 dimensional, dual polarization grating coupler.

    Based on a photonic crystal with a unit cell that is usually an ellipse,
    a rectangle or a circle.

    Args:
        unit_cell: component describing the unit cell of the photonic crystal.
        period_x: spacing between unit cells in the x direction [um].
        period_y: spacing between unit cells in the y direction [um].
        x_span: full x span of the photonic crystal.
        y_span: full y span of the photonic crystal.
        length_taper: taper length [um].
        width_taper: width of the taper at the grating coupler side [um].
        polarization: polarization of the grating coupler.
        wavelength: operation wavelength [um]
        taper: function to generate the tapers.
        base_layer: layer to draw over the whole photonic crystal
            (necessary if the unit cells are etched into a base layer).
        cross_section: for the routing waveguides.
        kwargs: cross_section settings.

    .. code::

        side view
                      fiber

                   /  /  /  /
                  /  /  /  /

                _|-|_|-|_|-|___  --> unit_cells
                   base_layer |
            o1  ______________|


        top view

                   -------------
               // | o   o   o  |
        o1 __ //  | o   o   o  |
              \\  | o   o   o  |
               \\ | o   o   o  |
                   -------------
                   \\         //
                    \\       //
                         |
                         o2

    """
    xs = gf.get_cross_section(cross_section, **kwargs)
    wg_width = xs.width
    layer = xs.layer

    c = Component()

    # ---------- First draw the grating coupler itself -----

    # Base layer
    _ = c << rectangle(
        size=(x_span, y_span), layer=base_layer, centered=True, port_type=None
    )

    # Photonic crystal
    num_x = int(np.floor(x_span / period_x))
    num_y = int(np.floor(y_span / period_y))
    x_start = -(num_x * period_x) / 2
    y_start = -(num_y * period_y) / 2
    x_end = -x_start
    y_end = -y_start
    x = np.linspace(x_start, x_end, num_x)
    y = np.linspace(y_start, y_end, num_y)
    xpos, ypos = np.meshgrid(x, y)

    for x, y in zip(xpos.flatten(), ypos.flatten()):
        un_cell = c << unit_cell()
        un_cell.x = gf.snap.snap_to_grid(x)
        un_cell.y = gf.snap.snap_to_grid(y)

    port_type = f"vertical_{polarization.lower()}"
    c.add_port(
        name=port_type,
        port_type=port_type,
        center=(0, 0),
        orientation=0,
        width=x_span,
        layer=layer,
    )
    c.info["polarization"] = polarization
    c.info["wavelength"] = wavelength

    # First taper
    taper1 = c << gf.get_component(
        taper,
        length=length_taper,
        width2=width_taper,
        width1=wg_width,
        layer=layer,
    )

    taper1.xmax = -x_span / 2
    taper1.y = 0
    c.add_port(port=taper1.ports["o1"], name="o1")

    # Second taper
    taper2 = c << gf.get_component(
        taper,
        length=length_taper,
        width2=width_taper,
        width1=wg_width,
        layer=layer,
    )
    taper2.rotate(90)

    taper2.x = 0
    taper2.ymax = -y_span / 2
    c.add_port(port=taper2.ports["o1"], name="o2")

    gf.asserts.grating_coupler(c)
    if xs.add_bbox:
        c = xs.add_bbox(c)
    if xs.add_pins:
        c = xs.add_pins(c)

    return c


if __name__ == "__main__":
    c = grating_coupler_dual_pol()
    c.show(show_ports=True)



from __future__ import annotations

from typing import Optional

import numpy as np


def mzi(
    wl: np.ndarray,
    neff: Optional[float],
    neff1: Optional[float] = None,
    neff2: Optional[float] = None,
    delta_length: Optional[float] = None,
    length1: Optional[float] = 0,
    length2: Optional[float] = None,
) -> np.ndarray:
    """Returns Frequency Domain Response of an MZI interferometer in linear units.

    Args:
        wl: wavelength in  um.
        neff: effective index.
        neff1: effective index branch 1.
        neff2: effective index branch 2.
        delta_length: length difference L2-L1.
        length1: length of branch 1.
        length2: length of branch 2.
    """
    k_0 = 2 * np.pi / wl
    length2 = length2 or length1 + delta_length
    delta_length = delta_length or np.abs(length2 - length1)
    neff1 = neff1 or neff
    neff2 = neff2 or neff

    E_out = 0.5 * (
        np.exp(1j * k_0 * neff1 * (length1 + delta_length))
        + np.exp(1j * k_0 * neff2 * length1)
    )
    return np.abs(E_out) ** 2


if __name__ == "__main__":
    import gdsfactory.simulation.gtidy3d as gt
    import matplotlib.pyplot as plt

    nm = 1e-3
    strip = gt.modes.Waveguide(
        wavelength=1.55,
        wg_width=500 * nm,
        wg_thickness=220 * nm,
        slab_thickness=0.0,
        ncore=gt.modes.si,
        nclad=gt.modes.sio2,
    )

    neff = 2.46  # Effective index of the waveguides
    wl0 = 1.55  # [μm] the wavelength at which neff and ng are defined
    wl = np.linspace(1.5, 1.6, 1000)  # [μm] Wavelengths to sweep over
    ngs = [4.182551, 4.169563, 4.172917]
    thicknesses = [210, 220, 230]

    length = 4e3
    dn = np.pi / length

    polyfit_TE1550SOI_220nm = np.array(
        [
            1.02478963e-09,
            -8.65556534e-08,
            3.32415694e-06,
            -7.68408985e-05,
            1.19282177e-03,
            -1.31366332e-02,
            1.05721429e-01,
            -6.31057637e-01,
            2.80689677e00,
            -9.26867694e00,
            2.24535191e01,
            -3.90664800e01,
            4.71899278e01,
            -3.74726005e01,
            1.77381560e01,
            -1.12666286e00,
        ]
    )

    def neff_w(wavelength):
        return np.poly1d(polyfit_TE1550SOI_220nm)(wavelength)

    w0 = 450 * nm
    dn1 = neff_w(w0 + 1 * nm / 2) - neff_w(w0 - 1 * nm / 2)
    dn5 = neff_w(w0 + 5 * nm / 2) - neff_w(w0 - 5 * nm / 2)
    dn10 = neff_w(w0 + 10 * nm / 2) - neff_w(w0 - 10 * nm / 2)

    pi_length1 = np.pi / dn1
    pi_length5 = np.pi / dn5
    pi_length10 = np.pi / dn10

    print(f"pi_length = {pi_length1:.0f}um for 1nm width variation")
    print(f"pi_length = {pi_length5:.0f}um for 5nm width variation")
    print(f"pi_length = {pi_length10:.0f}um for 10nm width variation")

    dn = dn1
    p = mzi(wl, neff=neff, neff1=neff + dn, neff2=neff + dn, delta_length=10)
    plt.plot(wl, p)
    plt.title("MZI")
    plt.xlabel("wavelength (um)")
    plt.ylabel("Power Transmission")
    plt.grid()
    plt.legend()
    plt.show()



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.bend_s import bend_s
from gdsfactory.components.straight import straight
from gdsfactory.typings import ComponentFactory, CrossSectionSpec, Floats, Optional


@gf.cell
def spiral_racetrack(
    min_radius: float = 5,
    straight_length: float = 10.0,
    spacings: Floats = (2, 2, 3, 3, 2, 2),
    straight_factory: ComponentFactory = straight,
    bend_factory: ComponentFactory = bend_euler,
    bend_s_factory: ComponentFactory = bend_s,
    cross_section: CrossSectionSpec = "strip",
    cross_section_s: CrossSectionSpec = None,
    n_bend_points: Optional[int] = None,
    with_inner_ports: bool = False,
) -> Component:
    """Returns Racetrack-Spiral.

    Args:
        min_radius: smallest radius in um.
        straight_length: length of the straight segments in um.
        spacings: space between the center of neighboring waveguides in um.
        straight_factory: factory to generate the straight segments.
        bend_factory: factory to generate the bend segments.
        bend_s_factory: factory to generate the s-bend segments.
        cross_section: cross-section of the waveguides.
        n_bend_points: optional bend points.
        with_inner_ports: if True, will build the spiral, but expose the inner ports where the S-bend would be.
    """
    c = gf.Component()

    if with_inner_ports:
        bend_s_component = bend_s_factory(
            (straight_length, -min_radius * 2 + 1 * spacings[0]),
            cross_section=cross_section_s if cross_section_s else cross_section,
            **({"nb_points": n_bend_points} if n_bend_points else {}),
        )
        bend_s = type("obj", (object,), {"ports": bend_s_component.ports})
        c.info["length"] = 0
        c.add_port(
            "o3",
            center=bend_s.ports["o1"].center,
            orientation=0,
            cross_section=bend_s.ports["o1"].cross_section,
        )
        c.add_port(
            "o4",
            center=bend_s.ports["o2"].center,
            orientation=180,
            cross_section=bend_s.ports["o2"].cross_section,
        )
    else:
        bend_s = c << bend_s_factory(
            (straight_length, -min_radius * 2 + 1 * spacings[0]),
            cross_section=cross_section_s if cross_section_s else cross_section,
            **({"nb_points": n_bend_points} if n_bend_points else {}),
        )
        c.info["length"] = bend_s.info["length"]

    ports = []
    for port in bend_s.ports.values():
        for i in range(len(spacings)):
            bend = c << bend_factory(
                angle=180,
                radius=min_radius + np.sum(spacings[:i]),
                p=0,
                cross_section=cross_section,
                **({"npoints": n_bend_points} if n_bend_points else {}),
            )
            bend.connect("o1", port)

            straight = c << straight_factory(
                straight_length, cross_section=cross_section
            )
            straight.connect("o1", bend.ports["o2"])
            port = straight.ports["o2"]

            c.info["length"] += bend.info["length"] + straight.info["length"]
        ports.append(port)

    c.add_port("o1", port=ports[0])
    c.add_port("o2", port=ports[1])
    return c


@gf.cell
def spiral_racetrack_heater_metal(
    min_radius: Optional[float] = None,
    straight_length: float = 30,
    spacing: float = 2,
    num: int = 8,
    straight_factory: ComponentFactory = straight,
    bend_factory: ComponentFactory = bend_euler,
    bend_s_factory: ComponentFactory = bend_s,
    waveguide_cross_section: CrossSectionSpec = "strip",
    heater_cross_section: CrossSectionSpec = "heater_metal",
) -> Component:
    """Returns spiral racetrack with a heater above.

    based on https://doi.org/10.1364/OL.400230 .

    Args:
        min_radius: smallest radius.
        straight_length: length of the straight segments.
        spacing: space between the center of neighboring waveguides.
        num: number.
        straight_factory: factory to generate the straight segments.
        bend_factory: factory to generate the bend segments.
        bend_s_factory: factory to generate the s-bend segments.
        waveguide_cross_section: cross-section of the waveguides.
        heater_cross_section: cross-section of the heater.
    """
    c = gf.Component()
    xs = gf.get_cross_section(waveguide_cross_section)
    min_radius = min_radius or xs.radius

    spiral = c << spiral_racetrack(
        min_radius,
        straight_length,
        (spacing,) * num,
        straight_factory,
        bend_factory,
        bend_s_factory,
        waveguide_cross_section,
    )

    heater_top = c << gf.components.straight(
        straight_length, cross_section=heater_cross_section
    )
    heater_top.connect("e1", spiral.ports["o1"].copy().rotate(180)).movey(
        spacing * num // 2
    )
    heater_bot = c << gf.components.straight(
        straight_length, cross_section=heater_cross_section
    )
    heater_bot.connect("e1", spiral.ports["o2"].copy().rotate(180)).movey(
        -spacing * num // 2
    )

    heater_bend = c << gf.components.bend_circular(
        angle=180,
        radius=min_radius + spacing * (num // 2 + 1),
        cross_section=heater_cross_section,
    )
    heater_bend.y = spiral.y
    heater_bend.x = spiral.x + min_radius + spacing * (num // 2 + 1)
    heater_top.connect("e1", heater_bend.ports["e1"])
    heater_bot.connect("e1", heater_bend.ports["e2"])

    c.add_ports(spiral.ports)
    c.add_port("e1", port=heater_bot["e2"])
    c.add_port("e2", port=heater_top["e2"])
    return c


@gf.cell
def spiral_racetrack_heater_doped(
    min_radius: Optional[float] = None,
    straight_length: float = 30,
    spacing: float = 2,
    num: int = 8,
    straight_factory: ComponentFactory = straight,
    bend_factory: ComponentFactory = bend_euler,
    bend_s_factory: ComponentFactory = bend_s,
    waveguide_cross_section: CrossSectionSpec = "strip",
    heater_cross_section: CrossSectionSpec = "npp",
) -> Component:
    """Returns spiral racetrack with a heater between the loops.

    based on https://doi.org/10.1364/OL.400230 but with the heater between the loops.

    Args:
        min_radius: smallest radius in um.
        straight_length: length of the straight segments in um.
        spacing: space between the center of neighboring waveguides in um.
        num: number.
        straight_factory: factory to generate the straight segments.
        bend_factory: factory to generate the bend segments.
        bend_s_factory: factory to generate the s-bend segments.
        waveguide_cross_section: cross-section of the waveguides.
        heater_cross_section: cross-section of the heater.
    """
    xs = gf.get_cross_section(waveguide_cross_section)
    min_radius = min_radius or xs.radius

    c = gf.Component()

    spiral = c << spiral_racetrack(
        min_radius=min_radius,
        straight_length=straight_length,
        spacings=(spacing,) * (num // 2)
        + (spacing + 1,) * 2
        + (spacing,) * (num // 2 - 2),
        straight_factory=straight_factory,
        bend_factory=bend_factory,
        bend_s_factory=bend_s_factory,
        cross_section=waveguide_cross_section,
    )

    heater_straight = gf.components.straight(
        straight_length, cross_section=heater_cross_section
    )

    heater_top = c << heater_straight
    heater_bot = c << heater_straight

    heater_bot.connect("e1", spiral.ports["o1"].copy().rotate(180)).movey(
        -spacing * (num // 2 - 1)
    )

    heater_top.connect("e1", spiral.ports["o2"].copy().rotate(180)).movey(
        spacing * (num // 2 - 1)
    )

    c.add_ports(spiral.ports)
    c.add_ports(prefix="top_", ports=heater_top.ports)
    c.add_ports(prefix="bot_", ports=heater_bot.ports)
    return c


if __name__ == "__main__":
    # heater = spiral_racetrack(
    #     min_radius=3.0, straight_length=30.0, spacings=(2, 2, 3, 3, 2, 2)
    # )
    # heater.show()

    # heater = spiral_racetrack_heater_metal(3, 30, 2, 5)
    # heater.show()

    # heater = spiral_racetrack_heater_doped(
    #     min_radius=3, straight_length=30, spacing=2, num=8
    # )
    # heater.show()
    c = spiral_racetrack(with_inner_ports=True)
    # c = spiral_racetrack_heater_doped()
    c.show(show_ports=True)



from gdsfactory.technology.layer_map import lyp_to_dataclass
from gdsfactory.technology.layer_stack import LayerLevel, LayerStack
from gdsfactory.technology.layer_views import LayerView, LayerViews
from gdsfactory.technology.simulation_settings import (
    SIMULATION_SETTINGS_LUMERICAL_FDTD,
    SimulationSettingsLumericalFdtd,
)

__all__ = [
    "LayerView",
    "LayerViews",
    "LayerLevel",
    "LayerStack",
    "lyp_to_dataclass",
    "SimulationSettingsLumericalFdtd",
    "SIMULATION_SETTINGS_LUMERICAL_FDTD",
]



"""Returns component with simulation markers."""
from __future__ import annotations

import warnings

import numpy as np

import gdsfactory as gf
from gdsfactory.add_pins import add_pin_rectangle
from gdsfactory.component import Component
from gdsfactory.components.bend_circular import bend_circular
from gdsfactory.pdk import get_layer_stack
from gdsfactory.technology import LayerLevel
from gdsfactory.typings import ComponentSpec, Layer, LayerSpec


@gf.cell
def add_simulation_markers(
    component: ComponentSpec = bend_circular,
    port_margin: float = 3,
    port_source_name: str = "o1",
    layer_source: Layer = (110, 0),
    layer_monitor: Layer = (101, 0),
    layer_label: LayerSpec = "TEXT",
    port_source_offset: float = 0.2,
) -> Component:
    r"""Returns new Component with simulation markers.

    Args:
        component: component spec.
        port_margin: from port edge in um.
        port_source_name: for input.
        layer_source: for port marker.
        layer_monitor: for port marker.
        layer_label: for labeling the ports.
        port_source_offset: distance from source to monitor in um.

    .. code::

         top view
              ________________________________
             |                               |
             | xmargin_left                  | port_extension
             |<------>          port_margin ||<-->
          ___|___________          _________||___
             |           \        /          |
             |            \      /           |
             |             ======            |
             |            /      \           |
          ___|___________/        \__________|___
             |   |                 <-------->|
             |   |ymargin_bot   xmargin_right|
             |   |                           |
             |___|___________________________|

        side view
              ________________________________
             |                     |         |
             |                     |         |
             |                   zmargin_top |
             |ymargin              |         |
             |<---> _____         _|___      |
             |     |     |       |     |     |
             |     |     |       |     |     |
             |     |_____|       |_____|     |
             |       |                       |
             |       |                       |
             |       |zmargin_bot            |
             |       |                       |
             |_______|_______________________|



    .. plot::
        :include-source:

        import gdsfactory as gf
        from gdsfactory.simulation.add_simulation_markers import add_simulation_markers

        c = gf.components.bend_circular()
        c = add_simulation_markers(c)
        c.plot()
    """
    c = gf.Component()
    component = gf.get_component(component)

    ref = c << component
    port_names = list(ref.ports.keys())

    layer_stack = get_layer_stack()

    if port_source_name not in port_names:
        warnings.warn(f"port_source_name={port_source_name!r} not in {port_names}")
        port_source = ref.get_ports_list()[0]
        port_source_name = port_source.name
        warnings.warn(f"Selecting port_source_name={port_source_name!r} instead.")

    assert isinstance(
        component, Component
    ), f"component needs to be a gf.Component, got Type {type(component)}"

    # Add port monitors
    for port_name in ref.ports.keys():
        port = ref.ports[port_name]
        add_pin_rectangle(c, port=port, port_margin=port_margin, layer=layer_monitor)
        layer_stack.layers["monitor"] = LayerLevel(
            layer=layer_monitor, thickness=2 * port_margin, zmin=-port_margin
        )

    # Add source
    port = ref.ports[port_source_name]
    angle_rad = np.radians(port.orientation)

    port = port.move_polar_copy(angle=angle_rad, d=port_source_offset)
    add_pin_rectangle(c, port=port, port_margin=port_margin, layer=layer_source)

    layer_stack.layers["source"] = LayerLevel(
        layer=layer_source, thickness=2 * port_margin, zmin=-port_margin
    )

    c.add_ports(component.ports)
    c.copy_child_info(component)
    return c


if __name__ == "__main__":
    # c = gf.components.coupler_ring()
    c = gf.components.mmi1x2()
    c = add_simulation_markers(c)
    c.show()
    scene = c.to_3d()
    scene.show()



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.pad import pad_array
from gdsfactory.typings import ComponentSpec, Coordinates, Float2, Optional


@cell
def add_fiducials(
    component: ComponentSpec = pad_array,
    gap: float = 50,
    left: Optional[ComponentSpec] = "cross",
    right: Optional[ComponentSpec] = "cross",
    top: Optional[ComponentSpec] = None,
    bottom: Optional[ComponentSpec] = None,
    offset: Float2 = (0, 0),
    **kwargs,
) -> Component:
    """Return component with fiducials.

    Args:
        component: component to add to the new component.
        gap: from component to fiducial edge.
        left: optional left fiducial.
        right: optional right fiducial.
        top: optional top fiducial.
        bottom: optional bottom fiducial.
        offset: component offset coordinate (x, y).
        kwargs: fiducial settings.
    """
    c = Component()
    component = gf.get_component(component, **kwargs)
    r = c << component
    r.move(offset)

    if left:
        x1 = c << gf.get_component(left)
        x1.xmax = r.xmin - gap
        c.add_ports(x1.ports, prefix="l")

    if right:
        x2 = c << gf.get_component(right)
        x2.xmin = r.xmax + gap
        c.add_ports(x2.ports, prefix="r")

    if top:
        y1 = c << gf.get_component(top)
        y1.ymin = r.ymax + gap
        c.add_ports(y1.ports, prefix="t")

    if bottom:
        y2 = c << gf.get_component(bottom)
        y2.ymax = r.ymin - gap
        c.add_ports(y2.ports, prefix="b")

    c.add_ports(r.ports)
    c.copy_child_info(component)
    return c


@cell
def add_fiducials_offsets(
    component: ComponentSpec = pad_array,
    fiducial: ComponentSpec = "cross",
    offsets: Coordinates = ((0, 100), (0, -100)),
) -> Component:
    """Returns new component with fiducials from a list of offsets.

    Args:
        component: add reference to component to the new Component.
        fiducial: function to return fiducial.
        offsets: list of offsets.
    """
    c = Component()
    component = gf.get_component(component)
    fiducial = gf.get_component(fiducial)
    r = c << component
    c.add_ports(r.ports)
    c.copy_child_info(component)

    for offset in offsets:
        f = c << fiducial
        f.move(offset)

    return c


if __name__ == "__main__":
    from gdsfactory.generic_tech import get_generic_pdk

    PDK = get_generic_pdk()
    PDK.activate()
    # c = add_fiducials(top='cross')
    c = add_fiducials_offsets()
    c.show(show_ports=True)



from __future__ import annotations

import numpy as np

import gdsfactory.simulation.gtidy3d as gt
from gdsfactory.simulation.gtidy3d.modes import Waveguide, group_index

nm = 1e-3
ncore = "si"
nclad = "sio2"


def test_neff_cached():
    c = Waveguide(
        wavelength=1.55,
        wg_width=0.5,
        wg_thickness=0.22,
        slab_thickness=0.0,
        ncore=ncore,
        nclad=nclad,
    )
    c.compute_modes()
    n0 = abs(c.neffs[0])
    assert np.isclose(n0, 2.46586, rtol=0.01), n0


def test_neff_no_cache():
    c = Waveguide(
        wavelength=1.55,
        wg_width=0.5,
        wg_thickness=0.22,
        slab_thickness=0.0,
        ncore=ncore,
        nclad=nclad,
        cache=False,
    )
    c.compute_modes()
    n0 = abs(c.neffs[0])
    assert np.isclose(n0, 2.46586, rtol=0.01), n0


def test_ng_no_cache():
    wg_settings = dict(
        wavelength=1.55,
        wg_width=0.5,
        wg_thickness=0.22,
        slab_thickness=0.0,
        ncore=ncore,
        nclad=nclad,
        cache=False,
    )
    ng = group_index(**wg_settings)
    assert np.isclose(ng, 4.169, rtol=0.01), ng


def test_sweep_width(dataframe_regression) -> None:
    df = gt.modes.sweep_width(
        width1=200 * nm,
        width2=1000 * nm,
        steps=1,
        wavelength=1.55,
        wg_thickness=220 * nm,
        slab_thickness=0 * nm,
        ncore=ncore,
        nclad=nclad,
        cache=False,
    )

    if dataframe_regression:
        dataframe_regression.check(df, default_tolerance=dict(atol=1e-3, rtol=1e-3))


if __name__ == "__main__":
    test_ng_no_cache()



from __future__ import annotations

import numpy as np

from gdsfactory.simulation.modes.find_modes import find_modes_waveguide


def test_find_modes_waveguide() -> None:
    modes = find_modes_waveguide(wg_width=0.45, resolution=20, cache=None)
    m1 = modes[1]
    m2 = modes[2]

    # neff1 = 2.3815558509779744
    # neff2 = 1.7749644180250004

    neff1 = 2.3494603726390664
    neff2 = 1.7030929743774146

    assert np.isclose(m1.neff, neff1), m1.neff
    assert np.isclose(m2.neff, neff2), m2.neff

    # Using cache
    modes = find_modes_waveguide(wg_width=0.45, resolution=20)
    m1 = modes[1]
    m2 = modes[2]

    assert np.isclose(m1.neff, neff1), m1.neff
    assert np.isclose(m2.neff, neff2), m2.neff


if __name__ == "__main__":
    test_find_modes_waveguide()
    # ms = get_mode_solver_rib(wg_width=0.45)
    # modes = find_neff(mode_solver=ms)
    # m1 = modes[1]
    # m2 = modes[2]
    # print(m1.neff)
    # print(m2.neff)
    # neff1 = 2.342628111145838
    # neff2 = 1.7286034634949181

    # assert np.isclose(m1.neff, neff1), (m1.neff, neff1)
    # assert np.isclose(m2.neff, neff2), (m2.neff, neff2)



from __future__ import annotations

from collections import OrderedDict
from typing import Dict, Optional, Tuple

import numpy as np
from scipy.interpolate import NearestNDInterpolator
from shapely.geometry import Polygon
from shapely.ops import unary_union

from gdsfactory.simulation.gmsh.mesh import mesh_from_polygons
from gdsfactory.simulation.gmsh.parse_component import (
    merge_by_material_func,
    process_buffers,
)
from gdsfactory.simulation.gmsh.parse_gds import cleanup_component
from gdsfactory.simulation.gmsh.parse_layerstack import (
    get_layers_at_z,
    order_layerstack,
)
from gdsfactory.technology import LayerStack
from gdsfactory.typings import ComponentOrReference


def xy_xsection_mesh(
    component: ComponentOrReference,
    z: float,
    layerstack: LayerStack,
    resolutions: Optional[Dict] = None,
    mesh_scaling_factor: float = 1.0,
    default_resolution_min: float = 0.01,
    default_resolution_max: float = 0.5,
    background_tag: Optional[str] = None,
    background_padding: Tuple[float, float, float, float] = (2.0, 2.0, 2.0, 2.0),
    filename: Optional[str] = None,
    global_meshsize_array: Optional[np.array] = None,
    global_meshsize_interpolant_func: Optional[callable] = NearestNDInterpolator,
    extra_shapes_dict: Optional[OrderedDict] = None,
    merge_by_material: Optional[bool] = False,
    round_tol: int = 3,
    simplify_tol: float = 1e-2,
):
    """Mesh xy cross-section of component at height z.

    Args:
        component (Component): gdsfactory component to mesh
        z (float): z-coordinate at which to sample the LayerStack
        layerstack (LayerStack): gdsfactory LayerStack to parse
        resolutions (Dict): Pairs {"layername": {"resolution": float, "distance": "float}} to roughly control mesh refinement
        mesh_scaling_factor (float): factor multiply mesh geometry by
        default_resolution_min (float): gmsh minimal edge length
        default_resolution_max (float): gmsh maximal edge length
        background_tag (str): name of the background layer to add (default: no background added)
        background_padding (Tuple): [xleft, ydown, xright, yup] distances to add to the components and to fill with background_tag
        filename (str, path): where to save the .msh file
        global_meshsize_array: np array [x,y,z,lc] to parametrize the mesh
        global_meshsize_interpolant_func: interpolating function for global_meshsize_array
        extra_shapes_dict: Optional[OrderedDict] = OrderedDict of {key: geo} with key a label and geo a shapely (Multi)Polygon or (Multi)LineString of extra shapes to override component
        merge_by_material: boolean, if True will merge polygons from layers with the same layer.material. Physical keys will be material in this case.
        round_tol: during gds --> mesh conversion cleanup, number of decimal points at which to round the gdsfactory/shapely points before introducing to gmsh
        simplify_tol: during gds --> mesh conversion cleanup, shapely "simplify" tolerance (make it so all points are at least separated by this amount)
    """
    # Fuse and cleanup polygons of same layer in case user overlapped them
    layer_polygons_dict = cleanup_component(
        component, layerstack, round_tol, simplify_tol
    )

    # GDS polygons to simulation polygons
    buffered_layer_polygons_dict, buffered_layerstack = process_buffers(
        layer_polygons_dict, layerstack
    )

    # Find layers present at this z-level
    layers = get_layers_at_z(buffered_layerstack, z)

    # Remove terminal layers and merge polygons
    layer_order = order_layerstack(buffered_layerstack)  # gds layers
    ordered_layers = [value for value in layer_order if value in set(layers)]
    shapes = OrderedDict() if extra_shapes_dict is None else extra_shapes_dict
    for layername in ordered_layers:
        for simulation_layername, (
            gds_layername,
            _next_simulation_layername,
            this_layer_polygons,
            _next_layer_polygons,
        ) in buffered_layer_polygons_dict.items():
            if simulation_layername == layername:
                shapes[gds_layername] = this_layer_polygons

    # Add background polygon
    if background_tag is not None:
        bounds = unary_union(list(shapes.values())).bounds
        shapes[background_tag] = Polygon(
            [
                [bounds[0] - background_padding[0], bounds[1] - background_padding[1]],
                [bounds[0] - background_padding[0], bounds[3] + background_padding[3]],
                [bounds[2] + background_padding[2], bounds[3] + background_padding[3]],
                [bounds[2] + background_padding[2], bounds[1] - background_padding[1]],
            ]
        )

    # Merge by material
    if merge_by_material:
        shapes = merge_by_material_func(shapes, layerstack)

    # Mesh
    return mesh_from_polygons(
        shapes,
        resolutions=resolutions,
        mesh_scaling_factor=mesh_scaling_factor,
        filename=filename,
        default_resolution_min=default_resolution_min,
        default_resolution_max=default_resolution_max,
        global_meshsize_array=global_meshsize_array,
        global_meshsize_interpolant_func=global_meshsize_interpolant_func,
    )


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.component.Component()
    waveguide = c << gf.get_component(gf.components.straight_pin(length=10, taper=None))
    undercut = c << gf.get_component(
        gf.components.rectangle(
            size=(5.0, 5.0),
            layer="UNDERCUT",
            centered=True,
        )
    ).move(destination=[4, 0])
    c.show()

    from gdsfactory.pdk import get_layer_stack

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                "slab90",
                "core",
                "via_contact",
                "undercut",
                "box",
                "substrate",
                # "clad",
                "metal1",
            )
        }
    )

    resolutions = {}
    resolutions["core"] = {"resolution": 0.05, "distance": 0.1}
    resolutions["via_contact"] = {"resolution": 0.1, "distance": 0}

    geometry = xy_xsection_mesh(
        component=c,
        z=-6,
        layerstack=filtered_layerstack,
        resolutions=resolutions,
        # background_tag="Oxide",
        filename="mesh.msh",
    )
    # print(geometry)

    # import gmsh

    # gmsh.write("mesh.msh")
    # gmsh.clear()
    # geometry.__exit__()

    import meshio

    mesh_from_file = meshio.read("mesh.msh")

    def create_mesh(mesh, cell_type, prune_z=False):
        cells = mesh.get_cells_type(cell_type)
        cell_data = mesh.get_cell_data("gmsh:physical", cell_type)
        points = mesh.points
        return meshio.Mesh(
            points=points,
            cells={cell_type: cells},
            cell_data={"name_to_read": [cell_data]},
        )

    line_mesh = create_mesh(mesh_from_file, "line", prune_z=True)
    meshio.write("facet_mesh.xdmf", line_mesh)

    triangle_mesh = create_mesh(mesh_from_file, "triangle", prune_z=True)
    meshio.write("mesh.xdmf", triangle_mesh)

    # # for layer, polygons in heaters.get_polygons(by_spec=True).items():
    # #     print(layer, polygons)



if __name__ == "__main__":
    import phidl.device_layout as pd
    import phidl.geometry as pg

    nm = 1e-3

    c = pd.Device()
    r1 = c << pg.compass(size=(1, 1))
    r1.movey(0.5 * nm)

    r2 = c << pg.compass(size=(1, 1))
    r2.connect("E", r1.ports["W"])

    c.write_gds("a.gds")
    import gdsfactory as gf

    gf.show("a.gds")



"""You can import gdsfactory.as gf.

functions:
    - import_gds(): returns a Component from a GDS

classes:

    - Component
    - Port
    - TECH

modules:

    - c: components
    - routing

isort:skip_file
"""
from __future__ import annotations
from warnings import warn
from functools import partial
from toolz import compose
from gdsfactory.component_layout import Group
from gdsfactory.path import Path


# NOTE: import order matters. Only change the order if you know what you are doing
from gdsfactory.component import Component, ComponentReference
from gdsfactory.config import CONF, call_if_func, PATH
from gdsfactory.port import Port
from gdsfactory.cell import cell
from gdsfactory.cell import cell_without_validator
from gdsfactory.cell import clear_cache
from gdsfactory.show import show
from gdsfactory.read.import_gds import import_gds
from gdsfactory.cross_section import CrossSection, Section
from gdsfactory.component_layout import Label
from gdsfactory import decorators
from gdsfactory import cross_section
from gdsfactory import labels
from gdsfactory import asserts
from gdsfactory import components
from gdsfactory import routing
from gdsfactory import typings
from gdsfactory import path
from gdsfactory import snap
from gdsfactory import read
from gdsfactory import add_termination
from gdsfactory import functions
from gdsfactory import geometry
from gdsfactory import add_ports
from gdsfactory import write_cells
from gdsfactory import add_pins
from gdsfactory import technology

from gdsfactory.add_tapers import add_tapers
from gdsfactory.add_padding import (
    add_padding,
    add_padding_container,
    get_padding_points,
)
from gdsfactory.fill import fill_rectangle
from gdsfactory.pack import pack
from gdsfactory.grid import grid, grid_with_text
from gdsfactory.generic_tech import LAYER, LAYER_STACK, get_generic_pdk
from gdsfactory.pdk import (
    Pdk,
    get_component,
    get_cross_section,
    get_layer,
    get_active_pdk,
    get_cell,
    get_constant,
)
from gdsfactory.get_factories import get_cells
from gdsfactory.cross_section import get_cross_section_factories

c = components


def __getattr__(name):
    if name == "types":
        warn("gdsfactory.types has been renamed to gdsfactory.typings")
        return typings
    raise AttributeError(f"No module named {name}")


__all__ = (
    "CONF",
    "Component",
    "ComponentReference",
    "CrossSection",
    "Group",
    "LAYER",
    "LAYER_STACK",
    "Label",
    "Path",
    "Pdk",
    "Port",
    "Section",
    "add_padding",
    "add_padding_container",
    "add_pins",
    "add_ports",
    "add_tapers",
    "add_termination",
    "asserts",
    "c",
    "call_if_func",
    "cell",
    "cell_without_validator",
    "clear_cache",
    "components",
    "compose",
    "cross_section",
    "decorators",
    "fill_rectangle",
    "functions",
    "geometry",
    "get_active_pdk",
    "get_cell",
    "get_cells",
    "get_component",
    "get_constant",
    "get_cross_section",
    "get_cross_section_factories",
    "get_layer",
    "get_padding_points",
    "get_generic_pdk",
    "grid",
    "grid_with_text",
    "import_gds",
    "labels",
    "pack",
    "partial",
    "path",
    "read",
    "routing",
    "show",
    "snap",
    "typings",
    "technology",
    "write_cells",
    "PATH",
)
__version__ = "6.53.0"



"""Preprocessing involving only the LayerStack."""
from __future__ import annotations

import numpy as np

from gdsfactory.technology import LayerStack


def list_unique_layerstack_z(
    layerstack: LayerStack,
):
    """List all unique LayerStack z coordinates.

    Args:
        layerstack: LayerStack
    Returns:
        Sorted set of z-coordinates for this layerstack
    """
    thicknesses = [layer.thickness for layer in layerstack.layers.values()]
    zmins = [layer.zmin for layer in layerstack.layers.values()]
    zmaxs = [sum(value) for value in zip(zmins, thicknesses)]

    return sorted(set(zmins + zmaxs))


def map_unique_layerstack_z(
    layerstack: LayerStack,
):
    """Map unique LayerStack z coordinates to various layers.

    Args:
        layerstack: LayerStack
    Returns:
        Dict with layernames as keys and set of unique z-values where the layer is present
    """
    z_levels = list_unique_layerstack_z(layerstack)
    layer_dict = layerstack.to_dict()
    unique_z_dict = {}
    for layername, layer in layer_dict.items():
        zmin = layer["zmin"]
        zmax = layer["zmin"] + layer["thickness"]
        if zmax > zmin:
            unique_z_dict[layername] = {
                z for z in z_levels if (z >= zmin and z <= zmax)
            }
        else:
            unique_z_dict[layername] = {
                z for z in z_levels if (z >= zmax and z <= zmin)
            }

    return unique_z_dict


def get_layer_overlaps_z(layerstack: LayerStack):
    """Maps layers to unique LayerStack z coordinates.

    Args:
        layerstack: LayerStack
    Returns:
        Dict with unique z-positions as keys, and list of layernames as entries
    """
    z_grid = list_unique_layerstack_z(layerstack)
    unique_z_dict = map_unique_layerstack_z(layerstack)
    intersection_z_dict = {}
    for z in z_grid:
        current_layers = {
            layername for layername, layer_zs in unique_z_dict.items() if z in layer_zs
        }
        intersection_z_dict[z] = current_layers

    return intersection_z_dict


def get_layers_at_z(layerstack: LayerStack, z: float):
    """Returns layers present at a given z-position.

    Args:
        layerstack: LayerStack
    Returns:
        List of layers
    """
    intersection_z_dict = get_layer_overlaps_z(layerstack)
    all_zs = list_unique_layerstack_z(layerstack)
    if z < np.min(all_zs):
        raise ValueError("Requested z-value is below the minimum layerstack z")
    elif z > np.max(all_zs):
        raise ValueError("Requested z-value is above the minimum layerstack z")
    for z_unique in intersection_z_dict.keys():
        if z <= z_unique:
            return intersection_z_dict[z_unique]
    raise AssertionError("Could not find z-value in layerstack z-range.")


def order_layerstack(layerstack: LayerStack):
    """Orders layerstack according to mesh_order.

    Args:
        layerstack: LayerStack.

    Returns:
        List of layernames: layerlevels dicts sorted by their mesh_order.
    """
    layers = layerstack.to_dict()
    mesh_orders = [value["mesh_order"] for value in layers.values()]
    return [x for _, x in sorted(zip(mesh_orders, layers))]


if __name__ == "__main__":
    import gdsfactory as gf

    waveguide = gf.components.straight_pin(length=1, taper=None)
    waveguide.show()

    from gdsfactory.pdk import get_layer_stack

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k] for k in ("core", "via_contact", "slab90")
        }
    )

    ret = order_layerstack(
        filtered_layerstack,
    )
    print(ret)



"""Straight Doped PIN waveguide."""
from __future__ import annotations

from typing import Optional

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.taper import taper_strip_to_ridge
from gdsfactory.components.via_stack import via_stack_m1_m3
from gdsfactory.components.via_stack_slot import via_stack_slot_slab_m1
from gdsfactory.cross_section import pin, pn
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@gf.cell
def straight_pin_slot(
    length: float = 500.0,
    cross_section: CrossSectionSpec = pin,
    via_stack: Optional[ComponentSpec] = via_stack_m1_m3,
    via_stack_width: float = 10.0,
    via_stack_slab: Optional[ComponentSpec] = via_stack_slot_slab_m1,
    via_stack_slab_top: Optional[ComponentSpec] = None,
    via_stack_slab_bot: Optional[ComponentSpec] = None,
    via_stack_slab_width: Optional[float] = None,
    via_stack_spacing: float = 3.0,
    via_stack_slab_spacing: float = 2.0,
    taper: Optional[ComponentSpec] = taper_strip_to_ridge,
    **kwargs,
) -> Component:
    """Returns a PIN straight waveguide with slotted via.

    https://doi.org/10.1364/OE.26.029983

    500um length for PI phase shift
    https://ieeexplore.ieee.org/document/8268112

    to go beyond 2PI, you will need at least 1mm
    https://ieeexplore.ieee.org/document/8853396/

    Args:
        length: of the waveguide.
        cross_section: for the waveguide.
        via_stack: for via_stacking the metal.
        via_stack_width: in um.
        via_stack_slab: function for the component via_stacking the slab.
        via_stack_slab_top: Optional, defaults to via_stack_slab.
        via_stack_slab_bot: Optional, defaults to via_stack_slab.
        via_stack_slab_width: defaults to via_stack_width.
        via_stack_spacing: spacing between via_stacks.
        via_stack_slab_spacing: spacing between via_stacks slabs.
        taper: optional taper.
        kwargs: cross_section settings.
    """
    c = Component()
    if taper:
        taper = gf.get_component(taper)
        length -= 2 * taper.get_ports_xsize()

    wg = c << gf.components.straight(
        cross_section=cross_section,
        length=length,
        **kwargs,
    )

    via_stack_slab_width = via_stack_slab_width or via_stack_width
    via_stack_slab_spacing = via_stack_slab_spacing or via_stack_spacing

    if taper:
        t1 = c << taper
        t2 = c << taper
        t1.connect("o2", wg.ports["o1"])
        t2.connect("o2", wg.ports["o2"])
        c.add_port("o1", port=t1.ports["o1"])
        c.add_port("o2", port=t2.ports["o1"])

    else:
        c.add_ports(wg.get_ports_list())

    via_stack_length = length

    if via_stack:
        via_stack_top = c << via_stack(
            size=(via_stack_length, via_stack_width),
        )
        via_stack_bot = c << via_stack(
            size=(via_stack_length, via_stack_width),
        )

        via_stack_bot.x = wg.x
        via_stack_top.x = wg.x

        via_stack_top.ymin = +via_stack_spacing / 2
        via_stack_bot.ymax = -via_stack_spacing / 2
        c.add_ports(via_stack_bot.ports, prefix="bot_")
        c.add_ports(via_stack_top.ports, prefix="top_")

    via_stack_slab_top = via_stack_slab_top or via_stack_slab
    via_stack_slab_bot = via_stack_slab_bot or via_stack_slab

    if via_stack_slab_top:
        slot_top = c << via_stack_slab_top(
            size=(via_stack_length, via_stack_slab_width),
        )
        slot_top.x = wg.x
        slot_top.ymin = +via_stack_slab_spacing / 2

    if via_stack_slab_bot:
        slot_bot = c << via_stack_slab_bot(
            size=(via_stack_length, via_stack_slab_width),
        )
        slot_bot.x = wg.x
        slot_bot.ymax = -via_stack_slab_spacing / 2

    return c


straight_pn_slot = gf.partial(straight_pin_slot, cross_section=pn)

if __name__ == "__main__":
    c = straight_pin_slot(via_stack_width=4, via_stack_slab_width=3, length=50)
    c.show(show_ports=True)



from sqlalchemy import TIMESTAMP, Column, Float, ForeignKey, Integer, String, text
from sqlalchemy.dialects.mysql import TEXT
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
metadata = Base.metadata


class Process(Base):
    __tablename__ = "process"
    __table_args__ = {"comment": "This table holds all foundry process info."}

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    name = Column(String(200), nullable=False)
    process = Column(String(200), nullable=False)
    version = Column(String(50), nullable=False)
    type = Column(String(50))
    description = Column(String(200))


class Unit(Base):
    __tablename__ = "unit"
    __table_args__ = {
        "comment": "This table holds all units. A unit is here understood as definite magnitude of a quantity."
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    name = Column(String(200), nullable=False)
    quantity = Column(String(200), nullable=False)
    symbol = Column(String(50), nullable=False)
    description = Column(String(200))


class Wafer(Base):
    __tablename__ = "wafer"
    __table_args__ = {"comment": "This table holds the base definition of a wafer."}

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    serial_number = Column(String(200), nullable=False)
    name = Column(String(200))
    description = Column(String(200))


class ComputedResult(Base):
    __tablename__ = "computed_result"
    __table_args__ = {
        "comment": "This table holds all results obtained after computation/analysis of the raw results contained in the table result."
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    name = Column(String(200), nullable=False)
    type = Column(String(50), nullable=False)
    unit_id = Column(ForeignKey("unit.id"), index=True)
    domain_unit_id = Column(ForeignKey("unit.id"), index=True)
    value = Column(TEXT, nullable=False)
    domain = Column(TEXT)
    description = Column(String(200))

    domain_unit = relationship(
        "Unit", primaryjoin="ComputedResult.domain_unit_id == Unit.id"
    )
    unit = relationship("Unit", primaryjoin="ComputedResult.unit_id == Unit.id")


class Result(Base):
    __tablename__ = "result"
    __table_args__ = {"comment": "This table holds all results."}

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    name = Column(String(200), nullable=False)
    type = Column(String(50), nullable=False)
    unit_id = Column(ForeignKey("unit.id"), index=True)
    domain_unit_id = Column(ForeignKey("unit.id"), index=True)
    value = Column(TEXT, nullable=False)
    domain = Column(TEXT)
    description = Column(String(200))

    domain_unit = relationship("Unit", primaryjoin="Result.domain_unit_id == Unit.id")
    unit = relationship("Unit", primaryjoin="Result.unit_id == Unit.id")


class Reticle(Base):
    __tablename__ = "reticle"
    __table_args__ = {"comment": "This table holds the definition of a reticle."}

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    name = Column(String(200), nullable=False)
    position = Column(
        String(50), comment="Position of the reticle on the wafer. (ROW, COLUMN)"
    )
    size = Column(
        String(50),
        comment="The size of the reticle (X,Y) having the convention that -Å· points towards the notch/flat of the wafer.",
    )
    wafer_id = Column(ForeignKey("wafer.id"), nullable=False, index=True)
    description = Column(String(200))

    wafer = relationship("Wafer")


class ComputedResultSelfRelation(Base):
    __tablename__ = "computed_result_self_relation"
    __table_args__ = {
        "comment": "This table holds all computed results self relation. This is used to link computed results together"
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    computed_result1_id = Column(
        ForeignKey("computed_result.id"), nullable=False, index=True
    )
    computed_result2_id = Column(
        ForeignKey("computed_result.id"), nullable=False, index=True
    )

    computed_result1 = relationship(
        "ComputedResult",
        primaryjoin="ComputedResultSelfRelation.computed_result1_id == ComputedResult.id",
    )
    computed_result2 = relationship(
        "ComputedResult",
        primaryjoin="ComputedResultSelfRelation.computed_result2_id == ComputedResult.id",
    )


class Die(Base):
    __tablename__ = "die"
    __table_args__ = {"comment": "This table holds die definition."}

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    reticle_id = Column(ForeignKey("reticle.id"), nullable=False, index=True)
    name = Column(String(200), nullable=False)
    position = Column(String(50))
    size = Column(String(50))
    description = Column(String(200))

    reticle = relationship("Reticle")


class ResultComputedResultRelation(Base):
    __tablename__ = "result_computed_result_relation"
    __table_args__ = {
        "comment": "This table holds the relations in between the results and the computed results."
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    result_id = Column(ForeignKey("result.id"), nullable=False, index=True)
    computed_result_id = Column(
        ForeignKey("computed_result.id"), nullable=False, index=True
    )

    computed_result = relationship("ComputedResult")
    result = relationship("Result")


class ResultInfo(Base):
    __tablename__ = "result_info"
    __table_args__ = {
        "comment": "This table holds extra information about specific results."
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    name = Column(String(200), nullable=False)
    value = Column(String(200), nullable=False)
    result_id = Column(ForeignKey("result.id"), nullable=False, index=True)
    computed_result_id = Column(
        ForeignKey("computed_result.id"), nullable=False, index=True
    )
    unit_id = Column(ForeignKey("unit.id"), index=True)
    description = Column(String(200))

    computed_result = relationship("ComputedResult")
    result = relationship("Result")
    unit = relationship("Unit")


class ResultProcessRelation(Base):
    __tablename__ = "result_process_relation"
    __table_args__ = {
        "comment": "This table holds all results and simulation result relation."
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    result_id = Column(ForeignKey("result.id"), nullable=False, index=True)
    process_id = Column(ForeignKey("process.id"), nullable=False, index=True)

    process = relationship("Process")
    result = relationship("Result")


class ResultSelfRelation(Base):
    __tablename__ = "result_self_relation"
    __table_args__ = {
        "comment": "This table holds all results self relation. This is used to link results together"
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    result1_id = Column(ForeignKey("result.id"), nullable=False, index=True)
    result2_id = Column(ForeignKey("result.id"), nullable=False, index=True)

    result1 = relationship(
        "Result", primaryjoin="ResultSelfRelation.result1_id == Result.id"
    )
    result2 = relationship(
        "Result", primaryjoin="ResultSelfRelation.result2_id == Result.id"
    )


class Component(Base):
    __tablename__ = "component"
    __table_args__ = {"comment": "This table holds the definition of components."}

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    die_id = Column(ForeignKey("die.id"), nullable=False, index=True)
    name = Column(String(250), nullable=False)
    description = Column(String(200))

    die = relationship("Die")


class Port(Base):
    __tablename__ = "port"
    __table_args__ = {"comment": "This table holds all ports definition."}

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    component_id = Column(ForeignKey("component.id"), nullable=False, index=True)
    name = Column(String(200), server_default=text("''"))
    port_type = Column(String(200))
    position = Column(String(50), nullable=False)
    orientation = Column(Float(asdecimal=True), nullable=False)
    description = Column(String(200))

    component = relationship("Component")


class ComponentInfo(Base):
    __tablename__ = "component_info"
    __table_args__ = {
        "comment": "This table holds information for the component using name/value pairs with optional description."
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    component_id = Column(ForeignKey("component.id"), index=True)
    die_id = Column(ForeignKey("die.id"), index=True)
    port_id = Column(ForeignKey("port.id"), index=True)
    reticle_id = Column(ForeignKey("reticle.id"), index=True)
    wafer_id = Column(ForeignKey("wafer.id"), index=True)
    name = Column(String(200), nullable=False)
    value = Column(String(200), nullable=False)
    description = Column(String(200))

    component = relationship("Component")
    die = relationship("Die")
    port = relationship("Port")
    reticle = relationship("Reticle")
    wafer = relationship("Wafer")


class ResultComponentRelation(Base):
    __tablename__ = "result_component_relation"
    __table_args__ = {
        "comment": "This table holds the relations in between results and components."
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    result_id = Column(ForeignKey("result.id"), nullable=False, index=True)
    component_id = Column(ForeignKey("component.id"), index=True)
    die_id = Column(ForeignKey("die.id"), index=True)
    port_id = Column(ForeignKey("port.id"), index=True)
    reticle_id = Column(ForeignKey("reticle.id"), index=True)
    wafer_id = Column(ForeignKey("wafer.id"), index=True)

    component = relationship("Component")
    die = relationship("Die")
    port = relationship("Port")
    result = relationship("Result")
    reticle = relationship("Reticle")
    wafer = relationship("Wafer")


class RelationInfo(Base):
    __tablename__ = "relation_info"
    __table_args__ = {
        "comment": "This table holds extra information about specific relation."
    }

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    computed_result_self_relation_id = Column(
        ForeignKey("computed_result_self_relation.id"), index=True
    )
    result_self_relation_id = Column(ForeignKey("result_self_relation.id"), index=True)
    result_process_relation_id = Column(
        ForeignKey("result_process_relation.id"), index=True
    )
    result_component_relation_id = Column(
        ForeignKey("result_component_relation.id"), index=True
    )
    result_computed_result_relation_id = Column(
        ForeignKey("result_computed_result_relation.id"), index=True
    )
    name = Column(String(200), nullable=False)
    value = Column(String(200), nullable=False)
    description = Column(String(200))

    computed_result_self_relation = relationship("ComputedResultSelfRelation")
    result_component_relation = relationship("ResultComponentRelation")
    result_computed_result_relation = relationship("ResultComputedResultRelation")
    result_process_relation = relationship("ResultProcessRelation")
    result_self_relation = relationship("ResultSelfRelation")


if __name__ == "__main__":
    from sqlalchemy import create_engine
    from sqlalchemy.orm import Session

    import gdsfactory as gf

    engine = create_engine("sqlite:///database.db", echo=True, future=True)
    metadata.create_all(engine)

    c = gf.components.ring_single(radius=10)

    with Session(engine) as session:
        w1 = Wafer(name="12", serial_number="ABC")
        r1 = Reticle(name="sky1", wafer_id=w1.id, wafer=w1)
        d1 = Die(name="d00", reticle_id=r1.id, reticle=r1)
        c1 = Component(name=c.name, die_id=d1.id, die=d1)

        component_settings = []

        for key, value in c.settings.changed.items():
            s = ComponentInfo(component=c1, component_id=c1.id, name=key, value=value)
            component_settings.append(s)

        for port in c.ports.values():
            s = Port(
                component=c1,
                component_id=c1.id,
                port_type=port.port_type,
                name=port.name,
                orientation=port.orientation,
                position=port.center,
            )
            component_settings.append(s)

        session.add_all([w1, r1, d1, c1])
        session.add_all(component_settings)
        session.commit()



from __future__ import annotations

import gdsfactory.simulation.modes as gm


def test_neff_vs_width(dataframe_regression) -> None:
    df = gm.find_neff_vs_width(steps=1, resolution=10, cache=None)
    if dataframe_regression:
        dataframe_regression.check(df)
    else:
        print(df)


if __name__ == "__main__":
    test_neff_vs_width(None)



from __future__ import annotations

import re
from functools import partial
from itertools import combinations
from typing import Dict, Optional, Sequence, Tuple

import matplotlib.pyplot as plt
import numpy as np

import gdsfactory as gf


def _check_ports(sp: Dict[str, np.ndarray], ports: Sequence[str]):
    """Ensure ports exist in Sparameters."""
    for port in ports:
        if port not in sp:
            raise ValueError(f"Did not find port {port!r} in {list(sp.keys())}")


def plot_sparameters(
    sp: Dict[str, np.ndarray],
    logscale: bool = True,
    keys: Optional[Tuple[str, ...]] = None,
    with_simpler_input_keys: bool = False,
    with_simpler_labels: bool = True,
) -> None:
    """Plots Sparameters from a dict of np.ndarrays.

    Args:
        sp: Sparameters np.ndarray.
        logscale: plots 20*log10(S).
        keys: list of keys to plot, plots all by default.
        with_simpler_input_keys: You can use S12 keys instead of o1@0,o2@0.
        with_simpler_labels: uses S11, S12 in plot labels instead of o1@0,o2@0.

    .. plot::
        :include-source:

        import gdsfactory as gf
        import gdsfactory.simulation as sim

        sp = sim.get_sparameters_data_lumerical(component=gf.components.mmi1x2)
        sim.plot.plot_sparameters(sp, logscale=True)

    """
    w = sp["wavelengths"] * 1e3
    keys = keys or [key for key in sp if not key.lower().startswith("wav")]

    for key in keys:
        if with_simpler_input_keys:
            key = f"o{key[1]}@0,o{key[2]}@0"
            if key not in sp:
                raise ValueError(f"{key!r} not in {list(sp.keys())}")

        if with_simpler_labels and "o" in key and "@" in key:
            port_mode1_port_mode2 = key.split(",")
            if len(port_mode1_port_mode2) != 2:
                raise ValueError(f"{key!r} needs to be 'portin@mode,portout@mode'")
            port_mode1, port_mode2 = port_mode1_port_mode2
            port1, _mode1 = port_mode1.split("@")
            port2, _mode2 = port_mode2.split("@")
            alias = f"S{port1[1:]}{port2[1:]}"
        else:
            alias = key

        if key not in sp:
            raise ValueError(f"{key!r} not in {list(sp.keys())}")
        y = sp[key]
        y = 20 * np.log10(np.abs(y)) if logscale else np.abs(y) ** 2
        plt.plot(w, y, label=alias)
    plt.legend()
    plt.xlabel("wavelength (nm)")
    plt.ylabel("|S| (dB)") if logscale else plt.ylabel("$|S|^2$")
    plt.show()


def plot_sparameters_phase(
    sp: Dict[str, np.ndarray],
    logscale: bool = True,
    keys: Optional[Tuple[str, ...]] = None,
    with_simpler_input_keys: bool = False,
    with_simpler_labels: bool = True,
) -> None:
    w = sp["wavelengths"] * 1e3
    keys = keys or [key for key in sp if not key.lower().startswith("wav")]

    for key in keys:
        if with_simpler_input_keys:
            key = f"o{key[1]}@0,o{key[2]}@0"
            if key not in sp:
                raise ValueError(f"{key!r} not in {list(sp.keys())}")

        if with_simpler_labels and "o" in key and "@" in key:
            port_mode1_port_mode2 = key.split(",")
            if len(port_mode1_port_mode2) != 2:
                raise ValueError(f"{key!r} needs to be 'portin@mode,portout@mode'")
            port_mode1, port_mode2 = port_mode1_port_mode2
            port1, _mode1 = port_mode1.split("@")
            port2, _mode2 = port_mode2.split("@")
            alias = f"S{port1[1:]}{port2[1:]}"
        else:
            alias = key

        if key not in sp:
            raise ValueError(f"{key!r} not in {list(sp.keys())}")
        y = sp[key]
        y = np.angle(y)
        plt.plot(w, y, label=alias)
    plt.legend()
    plt.xlabel("wavelength (nm)")
    plt.ylabel("S (deg)")
    plt.show()


def plot_imbalance(
    sp: Dict[str, np.ndarray], ports: Sequence[str], ax: Optional[plt.Axes] = None
) -> None:
    """Plots imbalance in dB for coupler.
    The imbalance is always defined between two ports, so this function plots the
    imbalance between all unique port combinations.

    Args:
        sp: sparameters dict np.ndarray.
        ports: list of port name @ mode index. o1@0 is the fundamental mode for o1 port.
        ax: matplotlib axis object to draw into.

    """
    _check_ports(sp, ports)

    power = {port: np.abs(sp[port]) ** 2 for port in ports}
    x = sp["wavelengths"] * 1e3

    if ax is None:
        _, ax = plt.subplots()

    for p1, p2 in combinations(ports, 2):
        p1in, p1out = re.findall(r"\d+", p1)[::2]
        p2in, p2out = re.findall(r"\d+", p2)[::2]
        label = f"$S_{{{p1in}{p1out}}}, S_{{{p2in}{p2out}}}$"
        ax.plot(x, 10 * np.log10(1 - (power[p1] - power[p2])), label=label)

    ax.set_xlim((x[0], x[-1]))
    ax.set_xlabel("wavelength (nm)")
    ax.set_ylabel("imbalance (dB)")
    plt.legend()


def plot_loss(
    sp: Dict[str, np.ndarray], ports: Sequence[str], ax: Optional[plt.Axes] = None
) -> None:
    """Plots loss dB for coupler.

    Args:
        sp: sparameters dict np.ndarray.
        ports: list of port name @ mode index. o1@0 is the fundamental mode for o1 port.
        ax: matplotlib axis object to draw into.

    """
    _check_ports(sp, ports)

    power = {port: np.abs(sp[port]) ** 2 for port in ports}
    x = sp["wavelengths"] * 1e3

    if ax is None:
        _, ax = plt.subplots()

    for n, p in power.items():
        pin, pout = re.findall(r"\d+", n)[::2]
        ax.plot(x, 10 * np.log10(p), label=f"$|S_{{{pin}{pout}}}|^2$")
    if len(ports) > 1:
        ax.plot(x, 10 * np.log10(sum(power.values())), "k--", label="Total")
    ax.set_xlim((x[0], x[-1]))
    ax.set_xlabel("wavelength (nm)")
    ax.set_ylabel("excess loss (dB)")
    plt.legend()


def plot_reflection(
    sp: Dict[str, np.ndarray], ports: Sequence[str], ax: Optional[plt.Axes] = None
) -> None:
    """Plots reflection in dB for coupler.

    Args:
        sp: sparameters dict np.ndarray.
        ports: list of port name @ mode index. o1@0 is the fundamental mode for o1 port.
        ax: matplotlib axis object to draw into.

    """
    _check_ports(sp, ports)

    power = {port: np.abs(sp[port]) ** 2 for port in ports}
    x = sp["wavelengths"] * 1e3

    if ax is None:
        _, ax = plt.subplots()

    for n, p in power.items():
        pin, pout = re.findall(r"\d+", n)[::2]
        ax.plot(x, 10 * np.log10(p), label=f"$|S_{{{pin}{pout}}}|^2$")
    if len(ports) > 1:
        ax.plot(x, 10 * np.log10(sum(power.values())), "k--", label="Total")
    ax.set_xlim((x[0], x[-1]))
    ax.set_xlabel("wavelength (nm)")
    ax.set_ylabel("reflection (dB)")
    plt.legend()


plot_loss1x2 = partial(plot_loss, ports=["o1@0,o2@0", "o1@0,o3@0"])
plot_loss2x2 = partial(plot_loss, ports=["o1@0,o3@0", "o1@0,o4@0"])
plot_imbalance1x2 = partial(plot_imbalance, ports=["o1@0,o2@0", "o1@0,o3@0"])
plot_imbalance2x2 = partial(plot_imbalance, ports=["o1@0,o3@0", "o1@0,o4@0"])
plot_reflection1x2 = partial(plot_reflection, ports=["o1@0,o1@0"])
plot_reflection2x2 = partial(plot_reflection, ports=["o1@0,o1@0", "o2@0,o1@0"])

if __name__ == "__main__":
    import gdsfactory.simulation as sim

    sp = sim.get_sparameters_data_tidy3d(component=gf.components.mmi1x2)
    # plot_sparameters(sp, logscale=False, keys=["o1@0,o2@0"])
    # plot_sparameters(sp, logscale=False, keys=["S21"])
    # plt.show()



"""Sample of a reticle top level Component."""

from __future__ import annotations

import gdsfactory as gf


if __name__ == "__main__":
    option = 3

    gc = gf.c.grating_coupler_elliptical_tm()
    c = gf.c.mzi_phase_shifter_top_heater_metal(delta_length=40)
    c = gf.routing.add_fiber_single(
        c, get_input_label_text_function=None, grating_coupler=gc
    )
    c = c.rotate(-90)

    if option == 1:
        c = gf.routing.add_electrical_pads_top(
            c, decorator=gf.add_labels.add_labels_to_ports
        )
    elif option == 2:
        c = gf.routing.add_electrical_pads_shortest(
            c, decorator=gf.add_labels.add_labels_to_ports
        )
    elif option == 3:
        c = gf.routing.add_electrical_pads_top_dc(
            c, decorator=gf.add_labels.add_labels_to_ports
        )
    c.show(show_ports=False)



from json import JSONEncoder, JSONDecoder

import numpy as np


class GdsfactoryJSONEncoder(JSONEncoder):
    """Encoder extending base :class:`.JSONEncoder` capabilities by serialising.

    * NumPy arrays
    * (WIP) GDS data
    """

    def default(self, o):
        if isinstance(o, np.ndarray):
            return o.tolist()
        elif np.iscomplex(o):
            return {"__complex__": True, "real": np.real(o), "imag": np.imag(o)}
        elif isinstance(o, np.generic):
            return o.item()
        else:
            return JSONEncoder.default(self, o)
        # match obj:
        #     case np.ndarray:
        #         return obj.tolist()
        #     case np.generic:
        #         return obj.item()
        #     case _:
        #         return JSONEncoder.default(self, obj)


class GdsfactoryJSONDecoder(JSONDecoder):
    """TODO."""

    def __init__(self, **kwargs):
        JSONDecoder.__init__(self, object_hook=self.object_hook, **kwargs)

    def object_hook(self, dct):
        if "__complex__" in dct:
            return complex(dct["real"], dct["imag"])
        return dct



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.triangles import triangle
from gdsfactory.typings import ComponentSpec, Float2, LayerSpec

triangle_metal = gf.partial(triangle, layer="M3", xtop=2)


@gf.cell
def dicing_lane(
    size: Float2 = (50, 300),
    marker: ComponentSpec = triangle_metal,
    layer_dicing: LayerSpec = "DICING",
) -> Component:
    """Dicing lane with triangular markers on both sides.

    Args:
        size: (tuple) Width and height of rectangle.
        marker: function to generate the dicing lane markers.
        layer_dicing: Specific layer to put polygon geometry on.
    """
    c = Component()
    r = c << rectangle(size=size, layer=layer_dicing)

    m = gf.get_component(marker)

    mbr = c << m
    mbr.xmin = r.xmax

    mbl = c << m
    mbl.mirror()
    mbl.xmax = r.xmin

    mtr = c << m
    mtr.mirror()
    mtr.rotate(180)
    mtr.xmin = r.xmax
    mtr.ymax = r.ymax

    mtl = c << m
    mtl.rotate(180)
    mtl.xmax = r.xmin
    mtl.ymax = r.ymax
    return c


if __name__ == "__main__":
    c = dicing_lane()
    c.show(show_ports=True)



from __future__ import annotations

from gdsfactory.export.to_3d import to_3d
from gdsfactory.export.to_np import to_np
from gdsfactory.export.to_stl import to_stl

__all__ = ("to_3d", "to_stl", "to_np")



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.simulation.simphony.model_from_gdsfactory import (
    GDSFactorySimphonyWrapper,
)


def mmi1x2(**kwargs):
    r"""Return 1x2 MultiModeInterferometer Sparameter model.

    Keyword Args:
        width: input and output straight width.
        width_taper: interface between input straights and mmi region.
        length_taper: into the mmi region.
        length_mmi: in x direction.
        width_mmi: in y direction.
        gap_mmi:  gap between tapered wg.
        taper: taper function.
        straight: straight function.
        with_bbox: add rectangular box in cross_section
            bbox_layers and bbox_offsets to avoid DRC sharp edges.
        cross_section: specification (CrossSection, string or dict).
    .. code::

               length_mmi
                <------>
                ________
               |        |
               |         \__
               |          __  o2
            __/          /_ _ _ _
        o1  __          | _ _ _ _| gap_mmi
              \          \__
               |          __  o3
               |         /
               |________|

             <->
        length_taper

    .. plot::
      :include-source:

      import gdsfactory as gf
      c = gf.components.mmi1x2(width_mmi=2, length_mmi=2.8)
      c.plot()

    .. plot::
        :include-source:

        import gdsfactory.simulation.simphony as gs
        import gdsfactory.simulation.simphony.components as gc

        c = gc.mmi1x2()
        gs.plot_model(c)
    """
    return GDSFactorySimphonyWrapper(component=gf.components.mmi1x2)


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    from gdsfactory.simulation.simphony.plot_model import plot_model

    wav = np.linspace(1520, 1570, 1024) * 1e-9
    f = 3e8 / wav
    c = mmi1x2()

    # s = c.s_parameters(freqs=f)
    # plt.plot(wav, np.abs(s[:, 1] ** 2))
    # print(c.pins)
    # print(c.settings)

    plot_model(c)
    plt.show()



from __future__ import annotations

from gdsfactory.routing.get_bundle import (
    get_bundle,
    get_bundle_electrical,
    get_bundle_same_axis_no_grouping,
)
from gdsfactory.routing.get_bundle_from_steps import (
    get_bundle_from_steps,
    get_bundle_from_steps_electrical,
)
from gdsfactory.routing.get_bundle_from_waypoints import get_bundle_from_waypoints
from gdsfactory.routing.get_bundle_path_length_match import get_bundle_path_length_match

from gdsfactory.routing.all_angle import get_bundle_all_angle

routing_strategy = dict(
    get_bundle=get_bundle,
    get_bundle_electrical=get_bundle_electrical,
    get_bundle_path_length_match=get_bundle_path_length_match,
    get_bundle_same_axis_no_grouping=get_bundle_same_axis_no_grouping,
    get_bundle_from_waypoints=get_bundle_from_waypoints,
    get_bundle_from_steps=get_bundle_from_steps,
    get_bundle_from_steps_electrical=get_bundle_from_steps_electrical,
    get_bundle_all_angle=get_bundle_all_angle,
)



from __future__ import annotations

import pathlib

import gdsfactory as gf
from gdsfactory.geometry.write_drc import (
    rule_area,
    rule_density,
    rule_enclosing,
    rule_separation,
    rule_space,
    rule_width,
    write_drc_deck_macro,
)

if __name__ == "__main__":
    filepath = pathlib.Path(__file__).parent.absolute() / "generic.lydrc"

    rules = [
        rule_width(layer="WG", value=0.2),
        rule_space(layer="WG", value=0.2),
        rule_width(layer="M1", value=1),
        rule_width(layer="M2", value=2),
        rule_space(layer="M2", value=2),
        rule_separation(layer1="HEATER", layer2="M1", value=1.0),
        rule_enclosing(layer1="M1", layer2="VIAC", value=0.2),
        rule_area(layer="WG", min_area_um2=0.05),
        rule_density(
            layer="WG", layer_floorplan="FLOORPLAN", min_density=0.5, max_density=0.6
        ),
    ]

    drc_rule_deck = write_drc_deck_macro(
        rules=rules, layers=gf.LAYER, shortcut="Ctrl+Shift+D", filepath=filepath
    )



"""Stream GDS to Klayout. Requires gdsfactory KLayout integration."""

from __future__ import annotations

import json
import os
import socket
from pathlib import Path
from typing import Union, Optional


def show(
    gds_filename: Union[Path, str],
    keep_position: bool = True,
    technology: Optional[str] = None,
    port: int = 8082,
) -> None:
    """Show GDS in klayout.

    Args:
        gds_filename: to show.
        keep_position: keep position and active layers.
        technology: Name of the KLayout technology to use.
        port: klayout server port.
    """
    if not os.path.isfile(gds_filename):
        raise ValueError(f"{gds_filename} does not exist")
    data = {
        "gds": os.path.abspath(gds_filename),
        "keep_position": keep_position,
        "technology": technology,
    }
    data_string = json.dumps(data)
    try:
        conn = socket.create_connection(("127.0.0.1", port), timeout=1.0)
        data_string = data_string + "\n"
        data_string = (
            data_string.encode() if hasattr(data_string, "encode") else data_string
        )
        conn.sendall(data_string)
        conn.close()
    except OSError:
        pass


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.straight()
    # gdspath = c.write_gds()
    # show(gdspath)
    c.show()



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.grating_coupler_elliptical_trenches import grating_coupler_te
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@cell
def grating_coupler_loss_fiber_single(
    grating_coupler: ComponentSpec = grating_coupler_te,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    """Returns grating coupler test structure.

    for testing with single fiber input/output

    Args:
        grating_coupler: function.
        cross_section: spec.

    Keyword Args:
        layer_label: for test and measurement label.
        min_input_to_output_spacing: spacing from input to output fiber.
        max_y0_optical: None.
        get_input_labels_function: function to get input labels for grating couplers.
        optical_routing_type: None: autoselection, 0: no extension.
        get_input_label_text_function: for the grating couplers input label.
        get_input_label_text_loopback_function: for the loopbacks input label.
    """
    c = gf.Component()
    grating_coupler = gf.get_component(grating_coupler)

    c << gf.routing.add_fiber_single(
        component=gf.components.straight(cross_section=cross_section),
        grating_coupler=grating_coupler,
        cross_section=cross_section,
        with_loopback=False,
        component_name=grating_coupler.name,
        **kwargs,
    )

    c.copy_child_info(grating_coupler)
    return c


if __name__ == "__main__":
    xs_strip2 = gf.partial(gf.cross_section.strip, layer=(2, 0))
    c = grating_coupler_loss_fiber_single(
        min_input_to_output_spacing=300, cross_section=xs_strip2
    )
    c.show(show_ports=True)



from __future__ import annotations

from typing import Callable, Dict, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from meep import mpb
from pydantic import BaseModel
from scipy.interpolate import RectBivariateSpline

from gdsfactory.typings import Array

# cmap_default = 'viridis'
cmap_default = "RdBu"


class Mode(BaseModel):
    """Mode object.

    Args:
        mode_number: for the mode.
        wavelength: um for the mode.
        neff: effective index.
        ng: group index.
        fraction_te: from 0 to 1.
        fraction_tm: from 0 to 1.
        effective_area: in um2.
        E: field.
        H: field.
        eps: permittivity.
        y: width.
        z: thickness.

    """

    mode_number: int
    wavelength: float
    neff: float
    ng: Optional[float] = None
    fraction_te: Optional[float] = None
    fraction_tm: Optional[float] = None
    effective_area: Optional[float] = None
    E: Optional[Array[float]] = None
    H: Optional[Array[float]] = None
    eps: Optional[Array[float]] = None
    y: Optional[Array[float]] = None
    z: Optional[Array[float]] = None

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"Mode{self.mode_number}"

    def E_grid_interp(self, y_arr, z_arr, index):
        """Creates new attributes with scipy.interpolate.RectBivariateSpline objects that can be used to interpolate the field on a new regular grid.

        Args:
            y_grid (np.array): y values where to evaluate, in increasing array.
            z_grid (np.array): z values where to evaluate, in increasing array.
            index: 0: x, 1: y, 2: z

        """
        if index not in [0, 1, 2]:
            raise ValueError(f"index = {index} needs to be (0: x, 1: y, 2: z)")

        return np.flip(
            RectBivariateSpline(self.y, self.z, np.real(self.E[:, :, 0, index]))(
                y_arr, z_arr, grid=True
            )
            + (
                1j
                * RectBivariateSpline(self.y, self.z, np.imag(self.E[:, :, 0, index]))(
                    y_arr, z_arr, grid=True
                )
            )
        )

    def Ex_grid_interp(self, y_arr, z_arr):
        return self.E_grid_interp(y_arr=y_arr, z_arr=z_arr, index=0)

    def Ey_grid_interp(self, y_arr, z_arr):
        return self.E_grid_interp(y_arr=y_arr, z_arr=z_arr, index=1)

    def Ez_grid_interp(self, y_arr, z_arr):
        return self.E_grid_interp(y_arr=y_arr, z_arr=z_arr, index=2)

    def H_grid_interp(self, y_arr, z_arr, index=0):
        """Creates new attributes with scipy.interpolate.RectBivariateSpline objects that can be used to interpolate the field on a new regular grid.

        Args:
            y_grid (np.array): y values where to evaluate, in increasing array.
            z_grid (np.array): z values where to evaluate, in increasing array.
            index: 0: x, 1: y, 2: z

        """
        if index not in [0, 1, 2]:
            raise ValueError(f"index = {index} needs to be (0: x, 1: y, 2: z)")
        return np.flip(
            RectBivariateSpline(self.y, self.z, np.real(self.H[:, :, 0, index]))(
                y_arr, z_arr, grid=True
            )
            + (
                1j
                * RectBivariateSpline(self.y, self.z, np.imag(self.H[:, :, 0, index]))(
                    y_arr, z_arr, grid=True
                )
            )
        )

    def Hx_grid_interp(self, y_arr, z_arr):
        return self.H_grid_interp(y_arr=y_arr, z_arr=z_arr, index=0)

    def Hy_grid_interp(self, y_arr, z_arr):
        return self.H_grid_interp(y_arr=y_arr, z_arr=z_arr, index=1)

    def Hz_grid_interp(self, y_arr, z_arr):
        return self.H_grid_interp(y_arr=y_arr, z_arr=z_arr, index=2)

    def plot_eps(
        self,
        cmap: str = "binary",
        origin="lower",
        logscale: bool = False,
        show: bool = True,
    ) -> None:
        """Plot index profile."""
        plt.imshow(
            self.eps**0.5,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
        )
        plt.title("index profile")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_e(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
    ) -> None:
        """Plot Electric field module."""
        E = self.E / abs(max(self.E.min(), self.E.max(), key=abs)) if scale else self.E
        Eabs = np.sqrt(
            np.multiply(E[:, :, 0, 2], E[:, :, 0, 2])
            + np.multiply(E[:, :, 0, 1], E[:, :, 0, 1])
            + np.multiply(E[:, :, 0, 0], E[:, :, 0, 0])
        )
        ep = abs(Eabs)
        ep = 10 * np.log10(ep) if logscale else ep
        plt.imshow(
            ep.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=0 if scale else ep.min(),
            vmax=1 if scale else ep.max(),
        )
        plt.title("$|E|$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_ex(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ) -> None:
        E = self.E / abs(max(self.E.min(), self.E.max(), key=abs)) if scale else self.E
        ex = E[:, :, 0, 0]
        ex = 10 * np.log10(np.abs(ex)) if logscale else operation(ex)
        plt.imshow(
            ex.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else ex.min(),
            vmax=1 if scale else ex.max(),
        )
        plt.title("$E_x$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_ey(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ) -> None:
        E = self.E / abs(max(self.E.min(), self.E.max(), key=abs)) if scale else self.E
        ey = E[:, :, 0, 1]
        ey = 10 * np.log10(np.abs(ey)) if logscale else operation(ey)
        plt.imshow(
            ey.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else ey.min(),
            vmax=1 if scale else ey.max(),
        )
        plt.title("$E_y$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_ez(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ) -> None:
        E = self.E / abs(max(self.E.min(), self.E.max(), key=abs)) if scale else self.E
        ez = E[:, :, 0, 2]
        ez = 10 * np.log10(ez) if logscale else operation(ez)
        plt.imshow(
            ez.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else ez.min(),
            vmax=1 if scale else ez.max(),
        )
        plt.title("$E_z$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_e_all(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ) -> None:
        plt.figure(figsize=(16, 10), dpi=100)

        plt.subplot(2, 3, 1)
        self.plot_ex(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 2)
        self.plot_ey(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 3)
        self.plot_ez(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 4)
        self.plot_e(show=False, scale=scale)

        plt.subplot(2, 3, 5)
        self.plot_eps(show=False)

        plt.tight_layout()
        plt.show()

    def plot_h(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
    ) -> None:
        H = self.H / abs(max(self.H.min(), self.H.max(), key=abs)) if scale else self.H
        Habs = np.sqrt(
            np.multiply(H[:, :, 0, 2], H[:, :, 0, 2])
            + np.multiply(H[:, :, 0, 1], H[:, :, 0, 1])
            + np.multiply(H[:, :, 0, 0], H[:, :, 0, 0])
        )
        hp = abs(Habs)
        hp = 10 * np.log10(hp) if logscale else hp
        plt.imshow(
            hp.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=0 if scale else hp.min(),
            vmax=1 if scale else hp.max(),
        )
        plt.title("$|H|$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_hx(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ) -> None:
        H = self.H / abs(max(self.H.min(), self.H.max(), key=abs)) if scale else self.H
        hx = H[:, :, 0, 0]
        hx = 10 * np.log10(np.abs(hx)) if logscale else operation(hx)
        plt.imshow(
            hx.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else hx.min(),
            vmax=1 if scale else hx.max(),
        )
        plt.title("$H_x$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_hy(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ) -> None:
        H = self.H / abs(max(self.H.min(), self.H.max(), key=abs)) if scale else self.H
        hy = H[:, :, 0, 1]
        hy = 10 * np.log10(np.abs(hy)) if logscale else operation(hy)
        plt.imshow(
            hy.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=-1 if scale else hy.min(),
            vmax=1 if scale else hy.max(),
        )
        plt.title("$H_y$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_hz(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ) -> None:
        H = self.H / abs(max(self.H.min(), self.H.max(), key=abs)) if scale else self.H
        hz = abs(H[:, :, 0, 2])
        hz = 10 * np.log10(hz) if logscale else operation(hz)
        plt.imshow(
            hz.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
            extent=[np.min(self.y), np.max(self.y), np.min(self.z), np.max(self.z)],
            vmin=0 if scale else hz.min(),
            vmax=1 if scale else hz.max(),
        )
        plt.title("$H_z$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_h_all(
        self,
        cmap: str = cmap_default,
        origin="lower",
        logscale: bool = False,
        show: bool = True,
        scale: bool = False,
        operation: Callable = np.real,
    ) -> None:
        plt.figure(figsize=(16, 10), dpi=100)

        plt.subplot(2, 3, 1)
        self.plot_hx(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 2)
        self.plot_hy(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 3)
        self.plot_hz(show=False, scale=scale, cmap=cmap, operation=operation)

        plt.subplot(2, 3, 4)
        self.plot_h(show=False, scale=scale)

        plt.subplot(2, 3, 5)
        self.plot_eps(show=False)

        plt.tight_layout()
        plt.show()


class Waveguide(BaseModel):
    """Waveguide Model.

    Args:
        wg_width: um
        wg_thickness: um
        slab_thickness: um
        ncore: core refractive index.
        nclad: cladding refractive index.
        sy: float
        sz: float
        resolution: int
        nmodes: int
        modes: Dict[Mode]

    """

    wg_width: float = 0.45
    wg_thickness: float = 0.22
    slab_thickness: float = 0
    ncore: float = 3.47
    nclad: float = 1.44
    sy: float = 2.0
    sz: float = 2.0
    resolution: int = 32
    nmodes: int = 4
    modes: Optional[Dict[int, Mode]] = None


class WavelengthSweep(BaseModel):
    wavelength: List[float]
    neff: Dict[int, List[float]]
    ng: Dict[int, List[float]]


ModeSolverFactory = Callable[..., mpb.ModeSolver]
ModeSolverOrFactory = Union[mpb.ModeSolver, ModeSolverFactory]


if __name__ == "__main__":
    import gdsfactory.simulation.modes as gm

    m = gm.find_modes_waveguide()
    # m[1].plot_e_all(operation=np.abs)
    # plt.show()
    # m[1].plot_h_all(operation=np.abs)
    # w = Waveguide()

    ys = np.linspace(-3, 3, 2000)
    zs = np.linspace(-1.5, 1.5, 2000)
    m[1].grid_interp()

    Ex_interp = m[1].Ex_grid_interp(ys, zs)
    Ey_interp = m[1].Ey_grid_interp(ys, zs)
    Ez_interp = m[1].Ez_grid_interp(ys, zs)

    plt.figure(figsize=(10, 8), dpi=100)

    plt.subplot(3, 2, 1)
    plt.title("As calculated")
    m[1].plot_ex(show=False, operation=np.abs, scale=False)

    plt.subplot(3, 2, 2)
    plt.title("Interp")
    plt.imshow(
        np.abs(Ex_interp).T,
        aspect="auto",
        extent=[np.min(ys), np.max(ys), np.min(zs), np.max(zs)],
    )
    plt.colorbar()

    plt.subplot(3, 2, 3)
    plt.title("As calculated")
    m[1].plot_ey(show=False, operation=np.abs, scale=False)

    plt.subplot(3, 2, 4)
    plt.title("Interp")
    plt.imshow(
        np.abs(Ey_interp).T,
        aspect="auto",
        extent=[np.min(ys), np.max(ys), np.min(zs), np.max(zs)],
    )
    plt.colorbar()

    plt.subplot(3, 2, 5)
    plt.title("As calculated")
    m[1].plot_ez(show=False, operation=np.abs, scale=False)

    plt.subplot(3, 2, 6)
    plt.title("Interp")
    plt.imshow(
        np.abs(Ez_interp).T,
        aspect="auto",
        extent=[np.min(ys), np.max(ys), np.min(zs), np.max(zs)],
    )
    plt.colorbar()

    plt.show()



from __future__ import annotations

from simphony import Model

import gdsfactory as gf
from gdsfactory.simulation.simphony.model_from_gdsfactory import (
    GDSFactorySimphonyWrapper,
)


def coupler_ring_fdtd(
    factory=gf.c.coupler_ring, width=0.5, length_x=4.0, gap=0.2, radius=5
) -> Model:
    r"""Return half ring model based on Lumerical 3D FDTD simulations.

    Args:
        c: gdsfactory component
        width:0.5
        gap: 0.2
        length_x: 4
        radius: 5

    .. code::

           N0            N1
           |             |
            \           /
             \         /
           ---=========---
        W0    length_x    E0


    """
    coupler = (
        factory(width=width, length_x=length_x, gap=gap, radius=radius)
        if callable(factory)
        else factory
    )
    return GDSFactorySimphonyWrapper(component=coupler)


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    import gdsfactory.simulation.simphony as gs

    wav = np.linspace(1520, 1570, 1024) * 1e-9
    c = coupler_ring_fdtd()
    wavelengths = np.linspace(1.5, 1.6) * 1e-6
    gs.plot_model(c, wavelengths=wavelengths)
    plt.show()

    # f = 3e8 / wav
    # s = c.s_parameters(freqs=f)
    # plt.plot(wav, np.abs(s[:, 1] ** 2))
    # print(c.pins)
    # plt.show()



from __future__ import annotations

import numpy as np

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.typings import LayerSpec


@cell
def optimal_90deg(
    width: float = 100,
    num_pts: int = 15,
    length_adjust: float = 1,
    layer: LayerSpec = (1, 0),
) -> Component:
    """Returns optimally-rounded 90 degree bend that is sharp on the outer corner.

    Args:
        width: Width of the ports on either side of the bend.
        num_pts: The number of points comprising the curved section of the bend.
        length_adjust: Adjusts the length of the non-curved portion of the bend.
        layer: Specific layer(s) to put polygon geometry on.

    Notes:
        Optimal structure from https://doi.org/10.1103/PhysRevB.84.174510
        Clem, J., & Berggren, K. (2011). Geometry-dependent critical currents in
        superconducting nanocircuits. Physical Review B, 84(17), 1–27.
    """
    D = Component()

    # Get points of ideal curve
    a = 2 * width
    v = np.logspace(-length_adjust, length_adjust, num_pts)
    xi = (
        a
        / 2.0
        * ((1 + 2 / np.pi * np.arcsinh(1 / v)) + 1j * (1 + 2 / np.pi * np.arcsinh(v)))
    )
    xpts = list(np.real(xi))
    ypts = list(np.imag(xi))

    # Add points for the rest of curve
    d = 2 * xpts[0]  # Farthest point out * 2, rounded to nearest 100
    xpts.append(width)
    ypts.append(d)
    xpts.append(0)
    ypts.append(d)
    xpts.append(0)
    ypts.append(0)
    xpts.append(d)
    ypts.append(0)
    xpts.append(d)
    ypts.append(width)
    xpts.append(xpts[0])
    ypts.append(ypts[0])

    D.add_polygon([xpts, ypts], layer=layer)

    D.add_port(name="e1", center=[a / 4, d], width=a / 2, orientation=90, layer=layer)
    D.add_port(name="e2", center=[d, a / 4], width=a / 2, orientation=0, layer=layer)
    return D


if __name__ == "__main__":
    c = optimal_90deg()
    c.show(show_ports=True)



from __future__ import annotations

import pathlib

import gdsfactory as gf
from gdsfactory.simulation.modes.find_coupling_vs_gap import (
    find_coupling,
    find_coupling_vs_gap,
)

PATH = pathlib.Path(__file__).parent.absolute() / "modes"

nm = 1e-3
wg_thickness = 220 * nm
nitride_thickness = 400 * nm
slab_thickness = 90 * nm
strip_width = 500 * nm
rib_width = 550 * nm


find_coupling_strip = gf.partial(
    find_coupling,
    wg_width=strip_width,
    wg_thickness=wg_thickness,
    slab_thickness=0.0,
)
find_coupling_rib = gf.partial(
    find_coupling,
    wg_width=rib_width,
    wg_thickness=wg_thickness,
    slab_thickness=slab_thickness,
)
find_coupling_nitride = gf.partial(
    find_coupling,
    wg_width=1.0,
    slab_thickness=0.0,
    ncore=2.0,
    wg_thickness=nitride_thickness,
    sz=4,
    ymargin=4,
)


find_coupling_vs_gap_strip = gf.partial(
    find_coupling_vs_gap,
)
find_coupling_vs_gap_rib = gf.partial(
    find_coupling_vs_gap,
    wg_width=rib_width,
    wg_thickness=wg_thickness,
    slab_thickness=slab_thickness,
)
find_coupling_vs_gap_nitride = gf.partial(
    find_coupling_vs_gap,
    wg_width=1.0,
    slab_thickness=0.0,
    ncore=2.0,
    wg_thickness=nitride_thickness,
    sz=4,
    ymargin=4,
    gap1=300 * nm,
    gap2=600 * nm,
)


if __name__ == "__main__":
    # import gdsfactory.simulation.modes as gm
    # df = find_coupling_vs_gap_strip()
    # gm.plot_coupling_vs_gap(df)

    # lc = find_coupling_vs_gap_strip()
    # lc = find_coupling_strip(gap=0.2)
    lc = find_coupling_nitride(gap=0.4)
    print(lc)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.pdk import get_layer_stack
from gdsfactory.simulation.gmsh.uz_xsection_mesh import uz_xsection_mesh
from gdsfactory.simulation.gmsh.xy_xsection_mesh import xy_xsection_mesh
from gdsfactory.simulation.gmsh.xyz_mesh import xyz_mesh
from gdsfactory.technology import LayerStack

PDK = gf.get_generic_pdk()
PDK.activate()


def test_gmsh_uz_xsection_mesh():
    waveguide = gf.components.straight_pin(length=10, taper=None)

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                "slab90",
                "core",
                "via_contact",
                # "metal2",
            )  # "slab90", "via_contact")#"via_contact") # "slab90", "core"
        }
    )

    resolutions = {
        "core": {"resolution": 0.05, "distance": 2},
        "slab90": {"resolution": 0.03, "distance": 1},
        "via_contact": {"resolution": 0.1, "distance": 1},
    }
    uz_xsection_mesh(
        waveguide,
        [(4, -15), (4, 15)],
        filtered_layerstack,
        resolutions=resolutions,
        background_tag="Oxide",
    )


def test_gmsh_xy_xsection_mesh():
    import gdsfactory as gf

    waveguide = gf.components.straight_pin(length=10, taper=None)
    waveguide.show()

    from gdsfactory.pdk import get_layer_stack

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                "slab90",
                "core",
                "via_contact",
            )
        }
    )

    resolutions = {
        "core": {"resolution": 0.05, "distance": 0.1},
        "via_contact": {"resolution": 0.1, "distance": 0},
    }
    xy_xsection_mesh(
        component=waveguide,
        z=0.09,
        layerstack=filtered_layerstack,
        resolutions=resolutions,
        background_tag="Oxide",
    )


def test_gmsh_xyz_holes():
    c = gf.component.Component()
    c << gf.get_component(gf.components.ring_crow)

    from gdsfactory.pdk import get_layer_stack

    filtered_layerstack = LayerStack(
        layers={k: get_layer_stack().layers[k] for k in ("core",)}
    )

    xyz_mesh(
        component=c,
        layerstack=filtered_layerstack,
        resolutions={},
        filename="mesh.msh",
        verbosity=False,
    )


if __name__ == "__main__":
    test_gmsh_xy_xsection_mesh()
    test_gmsh_uz_xsection_mesh()
    test_gmsh_xyz_holes()
